# cython: language_level=3, profile=True

import cython
from loguru import logger
import numpy as np
import trimesh

from libc.math cimport cos, sin, sqrt

from ... import name_patch
from ...cache import SharedCache


logger = name_patch(logger, __name__)

cdef dict ModelCache = {}
cdef int MaxModelCacheEntries = 4096
cdef double Tau = 6.283185307179586
cdef double RootHalf = sqrt(0.5)
cdef double DefaultSnapAngle = 0.05
cdef int DefaultSegments = 64
cdef Matrix44 IdentityTransform = Matrix44.__new__(Matrix44)


cdef class Model:
    def __hash__(self):
        return <Py_hash_t>(<void*>self)

    def __eq__(self, other):
        return self is other

    cdef bint is_constructed(self):
        return False

    cdef bint check_valid(self):
        raise NotImplementedError()

    cdef void build_trimesh_model(self):
        raise NotImplementedError()

    cdef tuple get_buffers(self, object glctx, dict objects):
        cdef str name = self.name
        if self.check_valid():
            if name in objects:
                return objects[name]
        else:
            self.build_trimesh_model()
        trimesh_model = self.trimesh_model
        if trimesh_model is None:
            if name in objects:
                del objects[name]
            return None, None
        while len(ModelCache) > MaxModelCacheEntries:
            dead_name = next(iter(ModelCache))
            del ModelCache[dead_name]
            if dead_name in objects:
                del objects[dead_name]
            logger.trace("Removed model {} from cache", dead_name)
        cdef tuple buffers
        cdef bint has_uv = trimesh_model.visual is not None and isinstance(trimesh_model.visual, trimesh.visual.texture.TextureVisuals)
        vertex_uvs = trimesh_model.visual.uv if has_uv else np.zeros((len(trimesh_model.vertices), 2))
        vertex_data = np.hstack((trimesh_model.vertices, trimesh_model.vertex_normals, vertex_uvs)).astype('f4')
        index_data = trimesh_model.faces.astype('i4')
        buffers = (glctx.buffer(vertex_data), glctx.buffer(index_data))
        logger.trace("Prepared model {} with {} vertices and {} faces", name, len(trimesh_model.vertices), len(trimesh_model.faces))
        objects[name] = buffers
        return buffers

    cdef tuple instance(self, Matrix44 model_matrix):
        return self, model_matrix

    cdef Model flatten(self):
        return FlattenedModel.get(self)

    cdef Model invert(self):
        return InvertedModel.get(self)

    cdef Model snap_edges(self, double snap_angle, double minimum_area):
        return SnappedEdgesModel.get(self, snap_angle, minimum_area)

    cdef Model transform(self, Matrix44 transform_matrix):
        return TransformedModel.get(self, transform_matrix)

    cdef Model slice(self, Vector origin, Vector normal):
        return SlicedModel.get(self, origin, normal)

    @staticmethod
    cdef Model intersect(list models):
        return BooleanOperationModel.get('intersection', models)

    @staticmethod
    cdef Model union(list models):
        return BooleanOperationModel.get('union', models)

    @staticmethod
    cdef Model difference(list models):
        return BooleanOperationModel.get('difference', models)

    @staticmethod
    cdef Model get_box(Node node):
        return Box.get(node)

    @staticmethod
    cdef Model get_sphere(Node node):
        return Sphere.get(node)

    @staticmethod
    cdef Model get_cylinder(Node node):
        return Cylinder.get(node)

    @staticmethod
    cdef Model get_cone(Node node):
        return Cone.get(node)

    @staticmethod
    cdef Model get_external(Node node):
        return ExternalModel.get(node)


cdef class ModelTransformer(Model):
    cdef Model original

    cdef bint check_valid(self):
        if not self.valid:
            return False
        if self.original.check_valid():
            return True
        self.valid = False
        return False


cdef class FlattenedModel(ModelTransformer):
    @staticmethod
    cdef FlattenedModel get(Model original):
        cdef str name = f'flat({original.name})'
        cdef FlattenedModel model = ModelCache.pop(name, None)
        if model is None:
            model = FlattenedModel.__new__(FlattenedModel)
            model.name = name
            model.original = original
        ModelCache[name] = model
        return model

    cdef void build_trimesh_model(self):
        if not self.original.check_valid():
            self.original.build_trimesh_model()
        trimesh_model = self.original.trimesh_model.copy()
        trimesh_model.unmerge_vertices()
        self.trimesh_model = trimesh_model
        self.valid = True


cdef class InvertedModel(ModelTransformer):
    @staticmethod
    cdef InvertedModel get(Model original):
        cdef str name = f'invert({original.name})'
        cdef InvertedModel model = ModelCache.pop(name, None)
        if model is None:
            model = InvertedModel.__new__(InvertedModel)
            model.name = name
            model.original = original
        ModelCache[name] = model
        return model

    cdef void build_trimesh_model(self):
        if not self.original.check_valid():
            self.original.build_trimesh_model()
        trimesh_model = self.original.trimesh_model.copy()
        trimesh_model.invert()
        self.trimesh_model = trimesh_model
        self.valid = True


cdef class SnappedEdgesModel(ModelTransformer):
    cdef double snap_angle
    cdef double minimum_area

    @staticmethod
    cdef SnappedEdgesModel get(Model original, double snap_angle, double minimum_area):
        cdef str name = 'snap(' + original.name
        if snap_angle != DefaultSnapAngle:
            name += f', {snap_angle:g}'
        if minimum_area:
            name += f', {minimum_area:g}'
        name += ')'
        cdef SnappedEdgesModel model = ModelCache.pop(name, None)
        if model is None:
            model = SnappedEdgesModel.__new__(SnappedEdgesModel)
            model.name = name
            model.original = original
            model.snap_angle = snap_angle
            model.minimum_area = minimum_area
        ModelCache[name] = model
        return model

    cdef void build_trimesh_model(self):
        if not self.original.check_valid():
            self.original.build_trimesh_model()
        if self.original.trimesh_model is not None:
            self.trimesh_model = trimesh.graph.smooth_shade(self.original.trimesh_model, angle=self.snap_angle*Tau,
                                                            facet_minarea=1/self.minimum_area if self.minimum_area else None)
        else:
            self.trimesh_model = None
        self.valid = True


cdef class TransformedModel(ModelTransformer):
    cdef Matrix44 transform_matrix

    @staticmethod
    cdef Model get(Model original, Matrix44 transform_matrix):
        if transform_matrix.eq(IdentityTransform):
            return original
        cdef str name = f'{original.name}@{hex(transform_matrix.hash(False))[3:]}'
        cdef TransformedModel model = ModelCache.pop(name, None)
        if model is None:
            model = TransformedModel.__new__(TransformedModel)
            model.name = name
            model.original = original
            model.transform_matrix = transform_matrix
        ModelCache[name] = model
        return model

    cdef tuple instance(self, Matrix44 model_matrix):
        return self.original, model_matrix.mmul(self.transform_matrix)

    cdef Model transform(self, Matrix44 transform_matrix):
        return TransformedModel.get(self.original, transform_matrix.mmul(self.transform_matrix))

    cdef void build_trimesh_model(self):
        if not self.original.check_valid():
            self.original.build_trimesh_model()
        if self.original.trimesh_model is not None:
            transform_array = np.array(self.transform_matrix, dtype='float64').reshape((4, 4)).transpose()
            trimesh_model = self.original.trimesh_model.copy().apply_transform(transform_array)
            self.trimesh_model = trimesh_model if len(trimesh_model.vertices) and len(trimesh_model.faces) else None
        else:
            self.trimesh_model = None
        self.valid = True


cdef class SlicedModel(ModelTransformer):
    cdef Vector origin
    cdef Vector normal

    @staticmethod
    cdef SlicedModel get(Model original, Vector origin, Vector normal):
        cdef str name = f'{original.name}/{hex(origin.hash(False) ^ normal.hash(False))[3:]}'
        cdef SlicedModel model = ModelCache.pop(name, None)
        if model is None:
            model = SlicedModel.__new__(SlicedModel)
            model.name = name
            model.original = original
            model.origin = origin
            model.normal = normal.normalize()
        ModelCache[name] = model
        return model

    cdef Model transform(self, Matrix44 transform_matrix):
        cdef Vector origin = transform_matrix.vmul(self.origin)
        cdef Vector normal = transform_matrix.inverse_transpose_matrix33().vmul(self.normal).normalize()
        return SlicedModel.get(self.original.transform(transform_matrix), origin, normal)

    cdef void build_trimesh_model(self):
        if not self.original.check_valid():
            self.original.build_trimesh_model()
        if self.original.trimesh_model is not None:
            trimesh_model = self.original.trimesh_model.copy().slice_plane(tuple(self.origin), tuple(self.normal.neg()), True)
            self.trimesh_model = trimesh_model if len(trimesh_model.vertices) and len(trimesh_model.faces) else None
        else:
            self.trimesh_model = None
        self.valid = True


cdef class BooleanOperationModel(Model):
    cdef str operation
    cdef list models

    cdef bint is_constructed(self):
        return True

    @staticmethod
    cdef Model get(str operation, list models):
        cdef Model child_model
        cdef list collected_models
        if operation == 'union':
            collected_models = []
            for child_model in models:
                if isinstance(child_model, BooleanOperationModel) and (<BooleanOperationModel>child_model).operation == 'union':
                    collected_models.extend((<BooleanOperationModel>child_model).models)
                else:
                    collected_models.append(child_model)
            models = collected_models
        cdef set existing = set()
        collected_models = []
        for child_model in models:
            if child_model is None:
                continue
            if child_model in existing:
                if operation == 'difference' and len(collected_models) and child_model is collected_models[0]:
                    return None
                continue
            existing.add(child_model)
            collected_models.append(child_model)
        models = collected_models
        if len(models) == 0:
            return None
        if len(models) == 1:
            return models[0]
        cdef str name = operation + '('
        cdef int i
        for i, child_model in enumerate(models):
            if i:
                name += ', '
            name += child_model.name
        name += ')'
        cdef BooleanOperationModel model = ModelCache.pop(name, None)
        if model is None:
            model = BooleanOperationModel.__new__(BooleanOperationModel)
            model.name = name
            model.operation = operation
            model.models = models
        ModelCache[name] = model
        return model

    cdef Model transform(self, Matrix44 transform_matrix):
        cdef Model model
        cdef list models = []
        for model in self.models:
            models.append(model.transform(transform_matrix))
        return BooleanOperationModel.get(self.operation, models)

    cdef Model slice(self, Vector origin, Vector normal):
        cdef Model model
        cdef list models = []
        cdef int i
        for i, model in enumerate(self.models):
            if i == 0 or self.operation == 'union':
                models.append(model.slice(origin, normal))
            else:
                models.append(model)
        return BooleanOperationModel.get(self.operation, models)

    cdef bint check_valid(self):
        if not self.valid:
            return False
        cdef Model model
        for model in self.models:
            if not model.check_valid():
                break
        else:
            return True
        self.valid = False
        return False

    cdef void build_trimesh_model(self):
        cdef list trimesh_models = []
        cdef Model model
        for model in self.models:
            if not model.check_valid():
                model.build_trimesh_model()
            if model.trimesh_model is None:
                continue
            trimesh_model = model.trimesh_model.copy()
            if not trimesh_model.is_watertight:
                trimesh_model.process(validate=True, merge_tex=True, merge_norm=True)
                if not trimesh_model.is_watertight and not trimesh_model.fill_holes():
                    logger.warning("Computing convex hull of: {}", model.name)
                    trimesh_model = trimesh_model.convex_hull
            trimesh_models.append(trimesh_model)
        if not trimesh_models:
            self.trimesh_model = None
        else:
            if self.operation == 'difference' and len(trimesh_models) > 2:
                union_models = trimesh.boolean.boolean_manifold(trimesh_models[1:], 'union')
                trimesh_model = trimesh.boolean.boolean_manifold([trimesh_models[0], union_models], 'difference')
            else:
                trimesh_model = trimesh.boolean.boolean_manifold(trimesh_models, self.operation)
            self.trimesh_model = trimesh_model if len(trimesh_model.vertices) and len(trimesh_model.faces) else None
        self.valid = True


cdef class PrimitiveModel(Model):
    cdef bint check_valid(self):
        return self.valid


cdef class Box(PrimitiveModel):
    Vertices = np.array([
        (-.5, -.5, +.5), (+.5, -.5, +.5), (+.5, +.5, +.5), (-.5, +.5, +.5),
        (-.5, +.5, +.5), (+.5, +.5, +.5), (+.5, +.5, -.5), (-.5, +.5, -.5),
        (+.5, +.5, +.5), (+.5, -.5, +.5), (+.5, -.5, -.5), (+.5, +.5, -.5),
        (+.5, +.5, -.5), (+.5, -.5, -.5), (-.5, -.5, -.5), (-.5, +.5, -.5),
        (-.5, +.5, -.5), (-.5, -.5, -.5), (-.5, -.5, +.5), (-.5, +.5, +.5),
        (-.5, -.5, -.5), (+.5, -.5, -.5), (+.5, -.5, +.5), (-.5, -.5, +.5),
    ], dtype='f4')
    VertexNormals = np.array([
        (0, 0, 1), (0, 0, 1), (0, 0, 1), (0, 0, 1),
        (0, 1, 0), (0, 1, 0), (0, 1, 0), (0, 1, 0),
        (1, 0, 0), (1, 0, 0), (1, 0, 0), (1, 0, 0),
        (0, 0, -1), (0, 0, -1), (0, 0, -1), (0, 0, -1),
        (-1, 0, 0), (-1, 0, 0), (-1, 0, 0), (-1, 0, 0),
        (0, -1, 0), (0, -1, 0), (0, -1, 0), (0, -1, 0),
    ], dtype='f4')
    VertexUV = np.array([
        (0, 0), (1/6, 0), (1/6, 1), (0, 1),
        (1/6, 0), (2/6, 0), (2/6, 1), (1/6, 1),
        (2/6, 0), (3/6, 0), (3/6, 1), (2/6, 1),
        (3/6, 0), (4/6, 0), (4/6, 1), (3/6, 1),
        (4/6, 0), (5/6, 0), (5/6, 1), (4/6, 1),
        (5/6, 0), (6/6, 0), (6/6, 1), (5/6, 1),
    ], dtype='f4')
    Faces = np.array([
        (0, 1, 2), (2, 3, 0),
        (4, 5, 6), (6, 7, 4),
        (8, 9, 10), (10, 11, 8),
        (12, 13, 14), (14, 15, 12),
        (16, 17, 18), (18, 19, 16),
        (20, 21, 22), (22, 23, 20),
    ], dtype='i4')

    @staticmethod
    cdef Box get(Node node):
        cdef str name = '!box'
        cdef Box model = ModelCache.pop(name, None)
        if model is None:
            model = Box.__new__(Box)
            model.name = name
            model.trimesh_model = None
        ModelCache[name] = model
        return model

    cdef void build_trimesh_model(self):
        visual = trimesh.visual.texture.TextureVisuals(uv=Box.VertexUV)
        self.trimesh_model = trimesh.base.Trimesh(vertices=Box.Vertices, vertex_normals=Box.VertexNormals, faces=Box.Faces, visual=visual)
        self.valid = True


cdef class Sphere(PrimitiveModel):
    cdef int segments

    @staticmethod
    cdef Sphere get(Node node):
        cdef int segments = max(4, node.get_int('segments', DefaultSegments))
        cdef str name = f'!sphere-{segments}' if segments != DefaultSegments else '!sphere'
        cdef Sphere model = ModelCache.pop(name, None)
        if model is None:
            model = Sphere.__new__(Sphere)
            model.name = name
            model.segments = segments
            model.trimesh_model = None
        ModelCache[name] = model
        return model

    @cython.cdivision(True)
    cdef void build_trimesh_model(self):
        cdef int ncols = self.segments, nrows = ncols//2, nvertices = (nrows+1)*(ncols+1), nfaces = (2+(nrows-2)*2)*ncols
        cdef object vertices_array = np.empty((nvertices, 3), dtype='f4')
        cdef float[:, :] vertices = vertices_array
        cdef object vertex_normals_array = np.empty((nvertices, 3), dtype='f4')
        cdef float[:, :] vertex_normals = vertex_normals_array
        cdef object vertex_uv_array = np.empty((nvertices, 2), dtype='f4')
        cdef float[:, :] vertex_uv = vertex_uv_array
        cdef object faces_array = np.empty((nfaces, 3), dtype='i4')
        cdef int[:, :] faces = faces_array
        cdef float x, y, z, r, th, u, v
        cdef int row, col, i=0, j=0
        for row in range(nrows + 1):
            v = <float>row/nrows
            if row == 0:
                r, z = 0, -1
            elif row == nrows:
                r, z = 0, 1
            else:
                th = Tau*(v-0.5)/2
                r, z = cos(th), sin(th)
            for col in range(ncols+1):
                u = (col+0.5)/ncols if row == 0 else ((col-0.5)/ncols if row == nrows else <float>col/ncols)
                if col == 0:
                    x, y = r, 0
                elif col == ncols:
                    x, y = r, 0
                else:
                    x, y = r*cos(Tau*u), r*sin(Tau*u)
                vertices[i, 0], vertices[i, 1], vertices[i, 2] = x, y, z
                vertex_normals[i, 0], vertex_normals[i, 1], vertex_normals[i, 2] = x, y, z
                vertex_uv[i, 0], vertex_uv[i, 1] = u, v
                if col < ncols and row < nrows:
                    if row < nrows-1:
                        faces[j, 0], faces[j, 1], faces[j, 2] = i, i+2+ncols, i+1+ncols
                        j += 1
                    if row > 0:
                        faces[j, 0], faces[j, 1], faces[j, 2] = i, i+1, i+2+ncols
                        j += 1
                i += 1
        visual = trimesh.visual.texture.TextureVisuals(uv=vertex_uv_array)
        self.trimesh_model = trimesh.base.Trimesh(vertices=vertices_array, vertex_normals=vertex_normals_array, faces=faces_array, visual=visual)
        self.valid = True


cdef class Cylinder(PrimitiveModel):
    cdef int segments

    @staticmethod
    cdef Cylinder get(Node node):
        cdef int segments = max(2, node.get_int('segments', DefaultSegments))
        cdef str name = f'!cylinder-{segments}' if segments != DefaultSegments else '!cylinder'
        cdef Cylinder model = ModelCache.pop(name, None)
        if model is None:
            model = Cylinder.__new__(Cylinder)
            model.name = name
            model.segments = segments
            model.trimesh_model = None
        ModelCache[name] = model
        return model

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef void build_trimesh_model(self):
        cdef int i, j, k, n = self.segments, m = (n+1)*6
        cdef object vertices_array = np.empty((m, 3), dtype='f4')
        cdef float[:, :] vertices = vertices_array
        cdef object vertex_normals_array = np.empty((m, 3), dtype='f4')
        cdef float[:, :] vertex_normals = vertex_normals_array
        cdef object vertex_uv_array = np.empty((m, 2), dtype='f4')
        cdef float[:, :] vertex_uv = vertex_uv_array
        cdef object faces_array = np.empty((n*4, 3), dtype='i4')
        cdef int[:, :] faces = faces_array
        cdef float x, y, th, u, u_
        for i in range(n+1):
            j = k = i * 6
            u = <float>i / n
            u_ = (i+0.5) / n
            if i == 0 or i == n:
                x, y = 1, 0
            else:
                th = Tau * u
                x, y = cos(th), sin(th)
            # bottom centre (k):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = 0, 0, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, -1
            vertex_uv[j, 0], vertex_uv[j, 1] = u_, 0
            j += 1
            # bottom edge (k+1):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, -1
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.25
            j += 1
            # side bottom (k+2):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = x, y, 0
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.25
            j += 1
            # side top (k+3):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, 0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = x, y, 0
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.75
            j += 1
            # top edge (k+4):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, 0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, 1
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.75
            j += 1
            # top centre (k+5):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = 0, 0, 0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, 1
            vertex_uv[j, 0], vertex_uv[j, 1] = u_, 1
            if i < n:
                j = i * 4
                # bottom face
                faces[j, 0], faces[j, 1], faces[j, 2] = k, k+1+6, k+1
                j += 1
                # side face 1
                faces[j, 0], faces[j, 1], faces[j, 2] = k+2+6, k+3, k+2
                j += 1
                # side face 2
                faces[j, 0], faces[j, 1], faces[j, 2] = k+3, k+2+6, k+3+6
                j += 1
                # top face
                faces[j, 0], faces[j, 1], faces[j, 2] = k+5, k+4, k+4+6
        visual = trimesh.visual.texture.TextureVisuals(uv=vertex_uv_array)
        self.trimesh_model = trimesh.base.Trimesh(vertices=vertices_array, vertex_normals=vertex_normals_array, faces=faces_array, visual=visual)
        self.valid = True


cdef class Cone(PrimitiveModel):
    cdef int segments

    @staticmethod
    cdef Cone get(Node node):
        cdef int segments = max(2, node.get_int('segments', DefaultSegments))
        cdef str name = f'!cone-{segments}' if segments != DefaultSegments else '!cone'
        cdef Cone model = ModelCache.pop(name, None)
        if model is None:
            model = Cone.__new__(Cone)
            model.name = name
            model.segments = segments
            model.trimesh_model = None
        ModelCache[name] = model
        return model

    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef void build_trimesh_model(self):
        cdef int i, j, k, n = self.segments, m = (n+1)*4
        cdef object vertices_array = np.empty((m, 3), dtype='f4')
        cdef float[:, :] vertices = vertices_array
        cdef object vertex_normals_array = np.empty((m, 3), dtype='f4')
        cdef float[:, :] vertex_normals = vertex_normals_array
        cdef object vertex_uv_array = np.empty((m, 2), dtype='f4')
        cdef float[:, :] vertex_uv = vertex_uv_array
        cdef object faces_array = np.empty((n*2, 3), dtype='i4')
        cdef int[:, :] faces = faces_array
        cdef float x, y, th, u, u_
        for i in range(n+1):
            j = k = i * 4
            u = <double>i / n
            u_ = (i+0.5) / n
            th_ = Tau * u_
            if i == 0 or i == n:
                x, y = 1, 0
            else:
                th = Tau * u
                x, y = cos(th), sin(th)
            # bottom centre (k):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = 0, 0, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, -1
            vertex_uv[j, 0], vertex_uv[j, 1] = u_, 0
            j += 1
            # bottom edge (k+1):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = 0, 0, -1
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.25
            j += 1
            # side bottom (k+2):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = x, y, -0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = x*RootHalf, y*RootHalf, RootHalf
            vertex_uv[j, 0], vertex_uv[j, 1] = u, 0.25
            j += 1
            # side top (k+3):
            vertices[j, 0], vertices[j, 1], vertices[j, 2] = 0, 0, 0.5
            vertex_normals[j, 0], vertex_normals[j, 1], vertex_normals[j, 2] = cos(th_)*RootHalf, sin(th_)*RootHalf, RootHalf
            vertex_uv[j, 0], vertex_uv[j, 1] = u_, 1
            if i < n:
                j = i * 2
                # bottom face
                faces[j, 0], faces[j, 1], faces[j, 2] = k, k+1+4, k+1
                j += 1
                # side face
                faces[j, 0], faces[j, 1], faces[j, 2] = k+3, k+2, k+2+4
        visual = trimesh.visual.texture.TextureVisuals(uv=vertex_uv_array)
        self.trimesh_model = trimesh.base.Trimesh(vertices=vertices_array, vertex_normals=vertex_normals_array, faces=faces_array, visual=visual)
        self.valid = True


cdef class ExternalModel(Model):
    cdef str filename

    @staticmethod
    cdef ExternalModel get(Node node):
        cdef str filename = node.get_str('filename', None)
        if not filename:
            return None
        cdef str name = filename
        cdef ExternalModel model = ModelCache.pop(name, None)
        if model is None:
            model = ExternalModel.__new__(ExternalModel)
            model.name = name
            model.filename = filename
            model.trimesh_model = None
        ModelCache[name] = model
        return model

    cdef bint check_valid(self):
        return self.trimesh_model is SharedCache[self.filename].read_trimesh_model()

    cdef void build_trimesh_model(self):
        self.trimesh_model = SharedCache[self.filename].read_trimesh_model()
        self.valid = True



"""
Flitter main entry point
"""

import argparse
import asyncio

from flitter import configure_logger, __version__
from .control import EngineController


def keyvalue(text):
    key, value = text.split('=', 1)
    values = value.split(';')
    for i in range(len(values)):
        try:
            values[i] = float(values[i])
        except ValueError:
            pass
    return key, values


def main():
    try:
        import setproctitle
    except ImportError:
        pass
    else:
        setproctitle.setproctitle('flitter')
    parser = argparse.ArgumentParser(description=f"Flitter language interpreter, version {__version__}")
    parser.set_defaults(level=None)
    levels = parser.add_mutually_exclusive_group()
    levels.add_argument('--trace', action='store_const', const='TRACE', dest='level', help="Trace logging")
    levels.add_argument('--debug', action='store_const', const='DEBUG', dest='level', help="Debug logging")
    levels.add_argument('--verbose', action='store_const', const='INFO', dest='level', help="Informational logging")
    levels.add_argument('--quiet', action='store_const', const='WARNING', dest='level', help="Only log warnings and errors")
    parser.add_argument('--profile', action='store_true', default=False, help="Run with profiling")
    parser.add_argument('--fps', type=int, default=60, help="Target framerate")
    parser.add_argument('--screen', type=int, default=0, help="Default screen number")
    parser.add_argument('--fullscreen', action='store_true', default=False, help="Default to full screen")
    parser.add_argument('--vsync', action='store_true', default=False, help="Default to window vsync")
    parser.add_argument('--state', type=str, help="State save/restore file")
    parser.add_argument('--autoreset', type=float, help="Auto-reset state on idle")
    parser.add_argument('--simplifystate', type=float, default=10, help="Simplify on state after stable period")
    parser.add_argument('--lockstep', action='store_true', default=False, help="Run clock in non-realtime mode")
    parser.add_argument('--define', '-D', action='append', default=[], type=keyvalue, dest='defines', help="Define name for evaluation")
    parser.add_argument('--vmstats', action='store_true', default=False, help="Report VM statistics")
    parser.add_argument('--runtime', type=float, help="Seconds to run for before exiting")
    parser.add_argument('--offscreen', action='store_true', default=False, help="Swap windows for offscreens")
    parser.add_argument('--gamma', type=float, default=1, help="Default window gamma correction")
    parser.add_argument('program', nargs='+', help="Program(s) to load")
    args = parser.parse_args()
    logger = configure_logger(args.level)
    logger.info("Flitter version {}", __version__)
    controller = EngineController(target_fps=args.fps, screen=args.screen, fullscreen=args.fullscreen, vsync=args.vsync,
                                  state_file=args.state, autoreset=args.autoreset, state_simplify_wait=args.simplifystate,
                                  realtime=not args.lockstep, defined_names=dict(args.defines), vm_stats=args.vmstats,
                                  run_time=args.runtime, offscreen=args.offscreen, window_gamma=args.gamma)
    for program in args.program:
        controller.load_page(program)

    try:
        if args.profile:
            import cProfile
            cProfile.runctx('asyncio.run(controller.run())', globals(), locals(), sort='tottime')
        else:
            asyncio.run(controller.run())
    except KeyboardInterrupt:
        logger.info("Exited on keyboard interrupt")
    except Exception:
        logger.error("Unexpected exception in flitter")
        raise
    finally:
        logger.complete()


if __name__ == '__main__':
    main()



"""
Ableton Push (and MIDI) constants
"""

import enum


class MIDI(enum.IntEnum):
    NOTE_OFF = 0x80
    NOTE_ON = 0x90
    POLYPHONIC_PRESSURE = 0xA0
    CONTROL_CHANGE = 0xB0
    CHANNEL_PRESSURE = 0xD0
    PITCH_BEND_CHANGE = 0xE0
    START_OF_SYSEX = 0xF0
    END_OF_SYSEX = 0xF7
    CLOCK = 0xF8
    START = 0xFA
    CONTINUE = 0xFB
    STOP = 0xFC
    ACTIVE = 0xFE
    RESET = 0xFF


class Command(enum.IntEnum):
    SET_COLOR_PALETTE_ENTRY = 0x03
    REAPPLY_COLOR_PALETTE = 0x05
    SET_LED_BRIGHTNESS = 0x06
    SET_DISPLAY_BRIGHTNESS = 0x08
    SET_MIDI_MODE = 0x0A
    SET_TOUCHSTRIP_CONFIG = 0x17
    SET_AFTERTOUCH_MODE = 0x1E


class Animation(enum.IntEnum):
    NONE = 0
    ONE_SHOT_SIXTH = 1
    ONE_SHOT_QUARTER = 2
    ONE_SHOT_HALF = 3
    ONE_SHOT_ONE = 4
    ONE_SHOT_TWO = 5
    PULSING_SIXTH = 6
    PULSING_QUARTER = 7
    PULSING_HALF = 8
    PULSING_ONE = 9
    PULSING_TWO = 10
    BLINKING_SIXTH = 11
    BLINKING_QUARTER = 12
    BLINKING_HALF = 13
    BLINKING_ONE = 14
    BLINKING_TWO = 15


class Note(enum.IntEnum):
    ENCODER_0 = 0
    ENCODER_1 = 1
    ENCODER_2 = 2
    ENCODER_3 = 3
    ENCODER_4 = 4
    ENCODER_5 = 5
    ENCODER_6 = 6
    ENCODER_7 = 7
    ENCODER_MASTER = 8
    ENCODER_METRONOME = 9
    ENCODER_TEMPO = 10
    TOUCH_STRIP = 12
    PAD_0_0 = 36
    PAD_0_1 = 37
    PAD_0_2 = 38
    PAD_0_3 = 39
    PAD_0_4 = 40
    PAD_0_5 = 41
    PAD_0_6 = 42
    PAD_0_7 = 43
    PAD_1_0 = 44
    PAD_1_1 = 45
    PAD_1_2 = 46
    PAD_1_3 = 47
    PAD_1_4 = 48
    PAD_1_5 = 49
    PAD_1_6 = 50
    PAD_1_7 = 51
    PAD_2_0 = 52
    PAD_2_1 = 53
    PAD_2_2 = 54
    PAD_2_3 = 55
    PAD_2_4 = 56
    PAD_2_5 = 57
    PAD_2_6 = 58
    PAD_2_7 = 59
    PAD_3_0 = 60
    PAD_3_1 = 61
    PAD_3_2 = 62
    PAD_3_3 = 63
    PAD_3_4 = 64
    PAD_3_5 = 65
    PAD_3_6 = 66
    PAD_3_7 = 67
    PAD_4_0 = 68
    PAD_4_1 = 69
    PAD_4_2 = 70
    PAD_4_3 = 71
    PAD_4_4 = 72
    PAD_4_5 = 73
    PAD_4_6 = 74
    PAD_4_7 = 75
    PAD_5_0 = 76
    PAD_5_1 = 77
    PAD_5_2 = 78
    PAD_5_3 = 79
    PAD_5_4 = 80
    PAD_5_5 = 81
    PAD_5_6 = 82
    PAD_5_7 = 83
    PAD_6_0 = 84
    PAD_6_1 = 85
    PAD_6_2 = 86
    PAD_6_3 = 87
    PAD_6_4 = 88
    PAD_6_5 = 89
    PAD_6_6 = 90
    PAD_6_7 = 91
    PAD_7_0 = 92
    PAD_7_1 = 93
    PAD_7_2 = 94
    PAD_7_3 = 95
    PAD_7_4 = 96
    PAD_7_5 = 97
    PAD_7_6 = 98
    PAD_7_7 = 99


class Control(enum.IntEnum):
    TAP_TEMPO = 3
    METRONOME = 9
    ENCODER_TEMPO = 14
    ENCODER_METRONOME = 15
    MENU_0_0 = 20
    MENU_0_1 = 21
    MENU_0_2 = 22
    MENU_0_3 = 23
    MENU_0_4 = 24
    MENU_0_5 = 25
    MENU_0_6 = 26
    MENU_0_7 = 27
    MASTER = 28
    STOP_CLIP = 29
    SETUP = 30
    LAYOUT = 31
    CONVERT = 35
    ONE_4 = 36
    ONE_4_T = 37
    ONE_8 = 38
    ONE_8_T = 39
    ONE_16 = 40
    ONE_16_T = 41
    ONE_32 = 42
    ONE_32_T = 43
    LEFT = 44
    RIGHT = 45
    UP = 46
    DOWN = 47
    SELECT = 48
    SHIFT = 49
    NOTE = 50
    SESSION = 51
    ADD_DEVICE = 52
    ADD_TRACK = 53
    OCTAVE_DOWN = 54
    OCTAVE_UP = 55
    REPEAT = 56
    ACCENT = 57
    SCALE = 58
    USER = 59
    MUTE = 60
    SOLO = 61
    PAGE_LEFT = 62
    PAGE_RIGHT = 63
    ENCODER_0 = 71
    ENCODER_1 = 72
    ENCODER_2 = 73
    ENCODER_3 = 74
    ENCODER_4 = 75
    ENCODER_5 = 76
    ENCODER_6 = 77
    ENCODER_7 = 78
    ENCODER_SETUP = 79
    PLAY = 85
    RECORD = 86
    NEW = 87
    DUPLICATE = 88
    AUTOMATE = 89
    FIXED_LENGTH = 90
    MENU_1_0 = 102
    MENU_1_1 = 103
    MENU_1_2 = 104
    MENU_1_3 = 105
    MENU_1_4 = 106
    MENU_1_5 = 107
    MENU_1_6 = 108
    MENU_1_7 = 109
    DEVICE = 110
    BROWSE = 111
    MIX = 112
    CLIP = 113
    QUANTIZE = 116
    DOUBLE_LOOP = 117
    DELETE = 118
    UNDO = 119


class Encoder(enum.IntEnum):
    ZERO = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    SETUP = 8
    METRONOME = 9
    TEMPO = 10


class TouchStripFlags(enum.IntFlag):
    PUSH2_CONTROL = 0b00000000
    HOST_CONTROL = 0b00000001
    SEND_VALUES = 0b00000000
    SEND_SYSEX = 0b00000010
    PITCH_BEND = 0b00000000
    MOD_WHEEL = 0b00000100
    LEDS_BAR = 0b00000000
    LEDS_POINT = 0b00001000
    BAR_BOTTOM = 0b00000000
    BAR_CENTER = 0b00010000
    NO_AUTO_RETURN = 0b00000000
    AUTO_RETURN = 0b00100000
    RETURN_BOTTOM = 0b00000000
    RETURN_CENTER = 0b01000000


BUTTONS = frozenset({Control.TAP_TEMPO, Control.METRONOME, Control.SETUP, Control.USER, Control.DELETE, Control.UNDO, Control.ADD_DEVICE,
                     Control.ADD_TRACK, Control.DEVICE, Control.BROWSE, Control.MIX, Control.CLIP, Control.MUTE, Control.SOLO, Control.STOP_CLIP,
                     Control.MASTER, Control.ONE_4, Control.ONE_4_T, Control.ONE_8, Control.ONE_8_T, Control.ONE_16, Control.ONE_16_T,
                     Control.ONE_32, Control.ONE_32_T, Control.CONVERT, Control.DOUBLE_LOOP, Control.QUANTIZE, Control.DUPLICATE, Control.NEW,
                     Control.FIXED_LENGTH, Control.AUTOMATE, Control.RECORD, Control.PLAY, Control.LEFT, Control.RIGHT, Control.UP, Control.DOWN,
                     Control.REPEAT, Control.ACCENT, Control.SCALE, Control.LAYOUT, Control.NOTE, Control.SESSION, Control.OCTAVE_DOWN, Control.OCTAVE_UP,
                     Control.PAGE_LEFT, Control.PAGE_RIGHT, Control.SELECT, Control.SHIFT,
                     Control.MENU_0_0, Control.MENU_0_1, Control.MENU_0_2, Control.MENU_0_3,
                     Control.MENU_0_4, Control.MENU_0_5, Control.MENU_0_6, Control.MENU_0_7,
                     Control.MENU_1_0, Control.MENU_1_1, Control.MENU_1_2, Control.MENU_1_3,
                     Control.MENU_1_4, Control.MENU_1_5, Control.MENU_1_6, Control.MENU_1_7})

COLOR_BUTTONS = frozenset({Control.MUTE, Control.SOLO, Control.STOP_CLIP, Control.AUTOMATE, Control.RECORD, Control.PLAY,
                           Control.ONE_4, Control.ONE_4_T, Control.ONE_8, Control.ONE_8_T,
                           Control.ONE_16, Control.ONE_16_T, Control.ONE_32, Control.ONE_32_T,
                           Control.MENU_0_0, Control.MENU_0_1, Control.MENU_0_2, Control.MENU_0_3,
                           Control.MENU_0_4, Control.MENU_0_5, Control.MENU_0_6, Control.MENU_0_7,
                           Control.MENU_1_0, Control.MENU_1_1, Control.MENU_1_2, Control.MENU_1_3,
                           Control.MENU_1_4, Control.MENU_1_5, Control.MENU_1_6, Control.MENU_1_7})

ENCODER_CONTROLS = {Control.ENCODER_0: Encoder.ZERO, Control.ENCODER_1: Encoder.ONE, Control.ENCODER_2: Encoder.TWO, Control.ENCODER_3: Encoder.THREE,
                    Control.ENCODER_4: Encoder.FOUR, Control.ENCODER_5: Encoder.FIVE, Control.ENCODER_6: Encoder.SIX, Control.ENCODER_7: Encoder.SEVEN,
                    Control.ENCODER_SETUP: Encoder.SETUP, Control.ENCODER_METRONOME: Encoder.METRONOME, Control.ENCODER_TEMPO: Encoder.TEMPO}



# cython: language_level=3, profile=False

from cpython cimport PyObject

from ..model cimport Vector, Context


cdef dict static_builtins
cdef dict dynamic_builtins
cdef dict builtins


cdef class VectorStack:
    cdef PyObject** vectors
    cdef int top
    cdef readonly int size

    cpdef VectorStack copy(self)
    cpdef void drop(self, int count=?)
    cpdef void push(self, Vector vector)
    cpdef Vector pop(self)
    cpdef tuple pop_tuple(self, int count)
    cpdef list pop_list(self, int count)
    cpdef dict pop_dict(self, tuple keys)
    cpdef Vector pop_composed(self, int count)
    cpdef Vector peek(self)
    cpdef Vector peek_at(self, int offset)
    cpdef void poke(self, Vector vector)
    cpdef void poke_at(self, int offset, Vector vector)


cdef class Program:
    cdef readonly list instructions
    cdef bint linked
    cdef readonly object path
    cdef readonly object top
    cdef readonly tuple initial_lnames
    cdef readonly VectorStack stack
    cdef readonly VectorStack lnames
    cdef int next_label

    cpdef void link(self)
    cpdef void optimize(self)
    cpdef int new_label(self)
    cpdef void dup(self)
    cpdef void drop(self, int count=?)
    cpdef void label(self, int label)
    cpdef void jump(self, int label)
    cpdef void branch_true(self, int label)
    cpdef void branch_false(self, int label)
    cpdef void pragma(self, str name)
    cpdef void import_(self, tuple names)
    cpdef void literal(self, value)
    cpdef void local_push(self, int count)
    cpdef void local_load(self, int offset)
    cpdef void local_drop(self, int count)
    cpdef void lookup(self)
    cpdef void lookup_literal(self, Vector value)
    cpdef void range(self)
    cpdef void neg(self)
    cpdef void pos(self)
    cpdef void ceil(self)
    cpdef void floor(self)
    cpdef void fract(self)
    cpdef void not_(self)
    cpdef void add(self)
    cpdef void sub(self)
    cpdef void mul(self)
    cpdef void mul_add(self)
    cpdef void truediv(self)
    cpdef void floordiv(self)
    cpdef void mod(self)
    cpdef void pow(self)
    cpdef void eq(self)
    cpdef void ne(self)
    cpdef void gt(self)
    cpdef void lt(self)
    cpdef void ge(self)
    cpdef void le(self)
    cpdef void xor(self)
    cpdef void slice(self)
    cpdef void slice_literal(self, Vector value)
    cpdef void call(self, int count, tuple names=?)
    cpdef void call_fast(self, function, int count)
    cpdef void tag(self, str name)
    cpdef void attribute(self, str name)
    cpdef void append(self, int count=?)
    cpdef void compose(self, int count)
    cpdef void begin_for(self)
    cpdef void next(self, int count, int label)
    cpdef void end_for_compose(self)
    cpdef void store_global(self, str name)
    cpdef void func(self, int label, str name, tuple parameters, int ncaptures=?)
    cpdef void exit(self)
    cdef void _execute(self, Context context, int pc, bint record_stats)



# cython: language_level=3, profile=True

"""
Flitter laser control
"""

import asyncio
import struct

import cython
from loguru import logger
from libc.math cimport sqrt, abs, round, cos, sin
import numpy as np
import usb.core

from .. import name_patch
from .. cimport model


logger = name_patch(logger, __name__)


cdef double TWO_PI = 2*np.pi


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:] travel(double distance, double accelleration, double start_speed, double end_speed):
    cdef double d0, t0, s, t1, tt, tm, t, tp, d
    cdef int i, n
    cdef double[:] ds
    d0 = abs(end_speed*end_speed - start_speed*start_speed) / (2*accelleration)
    t0 = abs(end_speed - start_speed) / accelleration
    s = max(start_speed, end_speed)
    t1 = (sqrt(s*s + 2*accelleration*(distance-d0) / 2) - s) / accelleration
    tt = t0 + 2*t1
    tm = t0 + t1 if end_speed > start_speed else t1
    n = max(<int>(round(tt)), 1)
    distances = np.zeros(n)
    ds = distances
    for i in range(1, n+1):
        t = i * tt / n
        tp = min(tm, t)
        d = start_speed*t + tp*tp*accelleration/2
        tp = max(0, t-tm)
        d += tp*tm*accelleration - tp*tp*accelleration/2
        ds[i-1] = d
    return distances

cdef double max_speed(double distance, double accelleration, double start_speed):
    return sqrt(2*distance*accelleration + start_speed * start_speed)

cdef double distance(tuple p0, tuple p1):
    cdef double x0, y0, x1, y1, dx, dy
    x0, y0 = p0
    x1, y1 = p1
    dx = x1 - x0
    dy = y1 - y0
    return sqrt(dx*dx + dy*dy)


cdef class LaserDriver:
    DEFAULT_SAMPLE_RATE = 30000
    DEFAULT_ACCELLERATION = 10
    DEFAULT_EPSILON = 0.0001

    cdef list _sample_bunches
    cdef int _sample_rate
    cdef double _accelleration, _epsilon

    @property
    def sample_rate(self):
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self._sample_rate = value

    @property
    def accelleration(self):
        return self._accelleration

    @accelleration.setter
    def accelleration(self, value):
        self._accelleration = value

    @property
    def epsilon(self):
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        self._epsilon = value

    def __init__(self, sample_rate=None, accelleration=None, epsilon=None):
        self._sample_bunches = [np.array([(0.5, 0.5, 0, 0, 0)])]
        self._sample_rate = sample_rate if sample_rate is not None else self.DEFAULT_SAMPLE_RATE
        self._accelleration = accelleration if accelleration is not None else self.DEFAULT_ACCELLERATION
        self._epsilon = epsilon if epsilon is not None else self.DEFAULT_EPSILON

    def close(self):
        pass

    async def flush(self):
        raise NotImplementedError()

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef draw_path(self, points, tuple color):
        cdef int i, j, n, m
        cdef double a, s, dv, dx, dy, d, x, y
        cdef double[:] accels, distances, speeds, lengths
        cdef double[:, :] samples, coords, normals
        cdef double r, g, b
        r = min(max(0, <double>(color[0])), 1)
        g = min(max(0, <double>(color[1])), 1)
        b = min(max(0, <double>(color[2])), 1)
        a = self._accelleration / self._sample_rate
        coords = np.vstack((self._sample_bunches[len(self._sample_bunches)-1][-1:, :2], points))
        n = len(coords)
        lengths = np.zeros(n-1)
        normals = np.zeros((n-1, 2))
        accels = np.zeros(n-1)
        i = 0
        for j in range(1, n):
            dx = coords[j, 0] - coords[i, 0]
            dy = coords[j, 1] - coords[i, 1]
            d = sqrt(dx*dx + dy*dy)
            if d < self._epsilon:
                n -= 1
                continue
            lengths[i] = d
            normals[i, 0] = dx / d
            normals[i, 1] = dy / d
            accels[i] = a / (max(abs(dx), abs(dy)) / d)
            i += 1
            if i != j:
                coords[i, 0] = coords[j, 0]
                coords[i, 1] = coords[j, 1]
        if n == 1:
            return
        speeds = np.zeros(n)
        for i in range(1, n-1):
            s = max_speed(lengths[i-1], accels[i-1], speeds[i-1])
            dv = max(abs(normals[i, 0] - normals[i-1, 0]), abs(normals[i, 1] - normals[i-1, 1]))
            speeds[i] = min(s, a / 2 / dv) if dv else s
        for i in range(n-2, 0, -1):
            speeds[i] = min(speeds[i], max_speed(lengths[i], accels[i], speeds[i+1]))
        for i in range(n-1):
            distances = travel(lengths[i], accels[i], speeds[i], speeds[i+1])
            m = len(distances)
            bunch = np.zeros((m, 5))
            samples = bunch
            for j in range(m):
                x = normals[i, 0] * distances[j] + coords[i, 0]
                y = normals[i, 1] * distances[j] + coords[i, 1]
                samples[j, 0] = min(max(0, x), 1)
                samples[j, 1] = min(max(0, y), 1)
                if i and 0 <= x <= 1 and 0 <= y <= 1:
                    samples[j, 2] = r
                    samples[j, 3] = g
                    samples[j, 4] = b
            self._sample_bunches.append(bunch)


cdef class LaserCubeDriver(LaserDriver):
    DEFAULT_SAMPLE_RATE = 50000  # samples/second
    DEFAULT_ACCELLERATION = 20   # sweeps/second^2
    DEFAULT_DAC_LAG = 2e-4       # seconds

    VENDOR_ID = 0x1fc9
    PRODUCT_ID = 0x04d8
    SET_ENABLED = 0x80
    GET_ENABLED = 0x81
    SET_DAC_RATE = 0x82
    GET_DAC_RATE = 0x83
    GET_MAXIMUM_DAC_RATE = 0x84
    GET_SAMPLE_ELEMENT_COUNT = 0x85
    GET_ISO_PACKET_SAMPLE_COUNT = 0x86
    GET_DAC_MINIMUM_VALUE = 0x87
    GET_DAC_MAXIMUM_VALUE = 0x88
    GET_RINGBUFFER_COUNT = 0x89
    GET_EMPTY_RINGBUFFER_COUNT = 0x8a
    GET_MAJOR_VERSION = 0x8b
    GET_MINOR_VERSION = 0x8c
    CLEAR_RINGBUFFER = 0x8d

    cdef object _device
    cdef object _control_out, _control_in, _data_out
    cdef int _dac_min, _dac_range, _max_dac_rate
    cdef double _dac_lag

    def __init__(self, id, dac_lag=None, **kwargs):
        super().__init__(**kwargs)
        for device in usb.core.find(idVendor=self.VENDOR_ID, idProduct=self.PRODUCT_ID, find_all=True):
            if id is None or device.serial_number == id:
                break
        else:
            raise ValueError("No device found")
        self._device = device
        self._device.set_configuration()
        self._device.set_interface_altsetting(1, 1)
        configuration = self._device.get_active_configuration()
        self._control_out, self._control_in = configuration[0, 0].endpoints()
        self._data_out, = configuration[1, 1].endpoints()
        self._dac_min = self._execute(self.GET_DAC_MINIMUM_VALUE, 'I')
        self._dac_range = self._execute(self.GET_DAC_MAXIMUM_VALUE, 'I') - self._dac_min
        self._max_dac_rate = self._execute(self.GET_MAXIMUM_DAC_RATE, 'I')
        self._dac_lag = dac_lag if dac_lag is not None else self.DEFAULT_DAC_LAG
        self._sample_rate = min(max(0, self._sample_rate), self._max_dac_rate)
        self._execute(self.SET_DAC_RATE, None, 'I', self._sample_rate)
        self._execute(self.SET_ENABLED, None, 'B', 1)
        major, minor = self._execute(self.GET_MAJOR_VERSION, 'I'), self._execute(self.GET_MINOR_VERSION, 'I')
        logger.info("Configured USB LaserCube '{} {}' (firmware {}.{})", self._device.manufacturer, self._device.product, major, minor)

    @property
    def sample_rate(self):
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        self._sample_rate = min(max(0, value), self._max_dac_rate)
        self._execute(self.SET_DAC_RATE, None, 'I', self._sample_rate)

    def close(self):
        self._execute(self.SET_ENABLED, None, 'B', 0)
        self._device = self._control_in = self._control_out = self._data_out = None

    def _execute(self, int command, str results_format, str values_format='', *values):
        self._control_out.write(struct.pack('<B' + values_format, command, *values))
        data = self._control_in.read(64)
        assert data[0] == command
        if not results_format:
            return None
        _, _, *results = struct.unpack_from("<BB" + results_format, data)
        return results if len(results) > 1 else results[0]

    @cython.boundscheck(False)
    async def flush(self):
        cdef int i, n, lag = max(1, <int>(round(self._sample_rate * self._dac_lag)))
        samples_array = np.vstack(self._sample_bunches)
        cdef double[:, :] samples = samples_array
        cdef unsigned short[:, :] output
        if len(samples) > lag:
            n = len(samples) - lag
            output_array = np.zeros((n, 4), dtype='uint16')
            output = output_array
            for i in range(n):
                output[i, 0] = <int>(round(samples[i, 2] * 255)) + (<int>(round(samples[i, 3] * 255)) << 8)
                output[i, 1] = <int>(round(samples[i, 4] * 255))
                output[i, 2] = <int>(round((1 - samples[i+lag, 0]) * self._dac_range)) + self._dac_min
                output[i, 3] = <int>(round((1 - samples[i+lag, 1]) * self._dac_range)) + self._dac_min
            logger.trace("Writing {} sample frame", len(output))
            await asyncio.to_thread(self._data_out.write, output_array.tobytes())
            self._sample_bunches = [samples_array[-lag:]]
        else:
            self._sample_bunches = [samples_array]


cdef class AffineTransform:
    cdef double v00, v01, v02
    cdef double v10, v11, v12
    cdef double v20, v21, v22

    @classmethod
    def identity(cls):
        cdef AffineTransform a = cls()
        a.v00 = a.v11 = a.v22 = 1
        return a

    @classmethod
    def translate(cls, double x, double y):
        cdef AffineTransform a = cls()
        a.v00 = a.v11 = a.v22 = 1
        a.v02 = x
        a.v12 = y
        return a

    @classmethod
    def scale(cls, double sx, double sy):
        cdef AffineTransform a = cls()
        a.v00 = sx
        a.v11 = sy
        a.v22 = 1
        return a

    @classmethod
    def rotate(cls, double th):
        cdef AffineTransform a = cls()
        cdef double sx = cos(th), sy = sin(th)
        a.v00 = sx
        a.v01 = -sy
        a.v10 = sy
        a.v11 = sx
        a.v22 = 1
        return a

    cpdef tuple t(self, double x, double y):
        return (x*self.v00 + y*self.v01 + self.v02, x*self.v10 + y*self.v11 + self.v12)

    def __matmul__(AffineTransform a, AffineTransform b):
        cdef AffineTransform c = AffineTransform()
        c.v00 = a.v00*b.v00 + a.v01*b.v10 + a.v02*b.v20
        c.v01 = a.v00*b.v01 + a.v01*b.v11 + a.v02*b.v21
        c.v02 = a.v00*b.v02 + a.v01*b.v12 + a.v02*b.v22
        c.v10 = a.v10*b.v00 + a.v11*b.v10 + a.v12*b.v20
        c.v11 = a.v10*b.v01 + a.v11*b.v11 + a.v12*b.v21
        c.v12 = a.v10*b.v02 + a.v11*b.v12 + a.v12*b.v22
        c.v20 = a.v20*b.v00 + a.v21*b.v10 + a.v22*b.v20
        c.v21 = a.v20*b.v01 + a.v21*b.v11 + a.v22*b.v21
        c.v22 = a.v20*b.v02 + a.v21*b.v12 + a.v22*b.v22
        return c


cdef class Laser:
    cdef LaserDriver driver

    def __init__(self, **kwargs):
        self.driver = None

    def purge(self):
        pass

    def destroy(self):
        if self.driver is not None:
            self.driver.close()
            self.driver = None

    async def update(self, engine, model.Node node, **kwargs):
        driver = node.get('driver', 1, str, '').lower()
        cls = {'lasercube': LaserCubeDriver}.get(driver)
        if cls is not None:
            id = node.get('id', 1, str)
            sample_rate = node.get('sample_rate', 1, int, cls.DEFAULT_SAMPLE_RATE)
            accelleration = node.get('accelleration', 1, float, cls.DEFAULT_ACCELLERATION)
            epsilon = node.get('epsilon', 1, float, cls.DEFAULT_EPSILON)
            if not isinstance(self.driver, cls):
                if self.driver is not None:
                    self.driver.close()
                self.driver = cls(id, sample_rate=sample_rate, accelleration=accelleration, epsilon=epsilon)
            else:
                if sample_rate != self.driver.sample_rate:
                    self.driver.sample_rate = sample_rate
                if accelleration != self.driver.accelleration:
                    self.driver.accelleration = accelleration
                if epsilon != self.driver.epsilon:
                    self.driver.epsilon = epsilon
            color = (0., 0., 0.)
            transform = AffineTransform.identity()
            paths = []
            self.collect_paths(node, paths, color, transform)
            if paths:
                self.draw_paths(paths)
            await self.driver.flush()
        elif self.driver is not None:
            self.driver.close()
            self.driver = None

    def draw_paths(self, list paths not None):
        cdef list points
        cdef tuple pair, color, last=(0.5, 0.5)
        cdef int i, nearest_i
        cdef double d, nearest_d=0
        while paths:
            for i in range(len(paths)):
                pair = paths[i]
                points = pair[0]
                d = distance(last, points[0])
                if i == 0 or d < nearest_d:
                    nearest_d = d
                    nearest_i = i
            pair = paths.pop(nearest_i)
            points = pair[0]
            color = pair[1]
            self.driver.draw_path(points, color)
            last = points[-1]

    @cython.cdivision(True)
    def collect_paths(self, model.Node node not None, list paths not None, tuple color not None, AffineTransform transform not None):
        cdef double x, y, sx, sy, th
        cdef int i, n
        cdef list path, points

        c = node.get('color', 3, float)
        if c is not None:
            color = tuple(c)

        if node.kind in ('laser', 'group'):
            for key in node.keys():
                if key == 'translate':
                    translate = node.get('translate', 2, float)
                    transform = transform @ AffineTransform.translate(*translate)
                elif key == 'scale':
                    scale = node.get('scale', 2, float)
                    transform = transform @ AffineTransform.scale(*scale)
                elif key == 'rotate':
                    rotate = node.get('rotate', 1, float)
                    transform = transform @ AffineTransform.rotate(TWO_PI * rotate)
            for child in node.children:
                self.collect_paths(child, paths, color, transform)

        elif node.kind == 'line':
            points = node.get('points', 0, float)
            n = len(points)
            if n >= 2:
                path = []
                for i in range(0, n, 2):
                    path.append(transform.t(points[i], points[i+1]))
                if node.get('close', 1, bool, False):
                    path.append(path[0])
                paths.append((path, color))

        elif node.kind == 'rect':
            size = node.get('size', 2, float)
            if size is not None:
                sx, sy = size
                x, y = node.get('point', 2, float, (0, 0))
                path = [transform.t(x, y), transform.t(x+sx, y), transform.t(x+sx, y+sy), transform.t(x, y+sy), transform.t(x, y)]
                paths.append((path, color))

        elif node.kind == 'ellipse':
            radius = node.get('radius', 2, float)
            if radius is not None:
                sx, sy = radius
                x, y = node.get('point', 2, float, (0, 0))
                n = node.get('segments', 1, int, 60)
                path = []
                for i in range(n+1):
                    th = TWO_PI * i/n
                    path.append(transform.t(x + sx*cos(th), y + sy*sin(th)))
                paths.append((path, color))


RENDERER_CLASS = Laser



# cython: language_level=3, profile=False

"""
Language abstract syntax tree

The tree supports reasonably sophisticated simplification that will reduce
the tree down to a "simpler" form by propogating constants. As this can result
in unrolling loops, simpler does not necessarily mean "smaller."
"""

from loguru import logger

from .. import name_patch
from ..model cimport Vector, Node, Context, StateDict, null_, true_, false_, minusone_
from .vm cimport Program, static_builtins, dynamic_builtins


logger = name_patch(logger, __name__)

cdef Literal NoOp = Literal(null_)


cdef Expression sequence_pack(list expressions):
    cdef Expression expr
    cdef Vector value
    cdef list vectors, remaining = []
    cdef bint has_let = False
    while expressions:
        expr = <Expression>expressions.pop(0)
        if isinstance(expr, Literal) and type((<Literal>expr).value) == Vector:
            vectors = []
            while isinstance(expr, Literal) and type((<Literal>expr).value) == Vector:
                value = (<Literal>expr).value
                if value.length:
                    vectors.append(value)
                if not expressions:
                    expr = None
                    break
                expr = <Expression>expressions.pop(0)
            if vectors:
                remaining.append(Literal(Vector._compose(vectors)))
        if expr is not None:
            if isinstance(expr, InlineSequence):
                expressions[:0] = (<InlineSequence>expr).expressions
                continue
            if isinstance(expr, (Let, Import, Function)):
                has_let = True
            remaining.append(expr)
    if len(remaining) == 0:
        return NoOp
    if has_let:
        for expr in remaining:
            if not isinstance(expr, Let):
                return Sequence(tuple(remaining))
        return NoOp
    if len(remaining) == 1:
        return remaining[0]
    return InlineSequence(tuple(remaining))


cdef class Expression:
    def compile(self, tuple initial_lnames=()):
        cdef Program program = Program.__new__(Program)
        program.initial_lnames = initial_lnames
        self._compile(program, list(initial_lnames))
        program.optimize()
        program.link()
        return program

    def simplify(self, StateDict state=None, dict static=None, dynamic=None):
        cdef dict context_vars = {}
        cdef str key
        if static is not None:
            for key, value in static.items():
                context_vars[key] = Vector._coerce(value)
        if dynamic is not None:
            for key in dynamic:
                context_vars[key] = None
        cdef Context context = Context(state=state, names=context_vars)
        cdef Expression expr
        try:
            expr = self._simplify(context)
        except Exception as exc:
            logger.opt(exception=exc).warning("Unable to simplify program")
            return self
        cdef str error
        for error in context.errors:
            logger.warning("Simplification error: {}", error)
        return expr

    cdef void _compile(self, Program program, list lnames):
        raise NotImplementedError()

    cdef Expression _simplify(self, Context context):
        raise NotImplementedError()


cdef class Top(Expression):
    cdef readonly tuple expressions

    def __init__(self, tuple expressions):
        self.expressions = expressions

    cdef void _compile(self, Program program, list lnames):
        cdef Expression expr
        cdef int m = 0
        for expr in self.expressions:
            expr._compile(program, lnames)
            if not isinstance(expr, (Let, Import, Function, Pragma, StoreGlobal)):
                m += 1
        if m:
            program.append(m)
        cdef str name
        m = len(program.initial_lnames)
        cdef int i, n=len(lnames)
        if n > m:
            for i in range(n-m):
                name = lnames[n-1-i]
                program.local_load(i)
                program.store_global(name)
            program.local_drop(n-m)

    cdef Expression _simplify(self, Context context):
        cdef list expressions = []
        cdef Expression expr
        cdef dict names = dict(context.names)
        for expr in self.expressions:
            expr = expr._simplify(context)
            if not isinstance(expr, Literal) or (<Literal>expr).value.length:
                expressions.append(expr)
        cdef str name
        cdef list bindings = []
        for name, value in context.names.items():
            if name not in names and value is not None and isinstance(value, Vector):
                bindings.append(Binding(name, Literal(value)))
        if bindings:
            expressions.append(StoreGlobal(tuple(bindings)))
        return Top(tuple(expressions))

    def __repr__(self):
        return f'Top({self.expressions!r})'


cdef class Pragma(Expression):
    cdef readonly str name
    cdef readonly Expression expr

    def __init__(self, str name, Expression expr):
        self.name = name
        self.expr = expr

    cdef void _compile(self, Program program, list lnames):
        self.expr._compile(program, lnames)
        program.pragma(self.name)

    cdef Expression _simplify(self, Context context):
        return Pragma(self.name, self.expr._simplify(context))

    def __repr__(self):
        return f'Pragma({self.name!r}, {self.expr!r})'


cdef class Import(Expression):
    cdef readonly tuple names
    cdef readonly Expression filename

    def __init__(self, tuple names, Expression filename):
        self.names = names
        self.filename = filename

    cdef void _compile(self, Program program, list lnames):
        self.filename._compile(program, lnames)
        program.import_(self.names)
        lnames.extend(self.names)

    cdef Expression _simplify(self, Context context):
        cdef str name
        for name in self.names:
            context.names[name] = None
        return Import(self.names, self.filename._simplify(context))

    def __repr__(self):
        return f'Import({self.names!r}, {self.filename!r})'


cdef class Sequence(Expression):
    cdef readonly tuple expressions

    def __init__(self, tuple expressions):
        self.expressions = expressions

    cdef void _compile(self, Program program, list lnames):
        cdef Expression expr
        cdef int n=len(lnames), m=0
        for expr in self.expressions:
            expr._compile(program, lnames)
            if not isinstance(expr, (Let, Import, Function, Pragma)):
                m += 1
        if len(lnames) > n:
            program.local_drop(len(lnames)-n)
            while len(lnames) > n:
                lnames.pop()
        if m > 1:
            program.compose(m)
        elif m == 0:
            program.literal(null_)

    cdef Expression _simplify(self, Context context):
        cdef list expressions = []
        cdef Expression expr
        cdef dict saved = context.names
        context.names = saved.copy()
        for expr in self.expressions:
            expressions.append(expr._simplify(context))
        context.names = saved
        return sequence_pack(expressions)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.expressions!r})'


cdef class InlineSequence(Sequence):
    cdef void _compile(self, Program program, list lnames):
        cdef Expression expr
        for expr in self.expressions:
            expr._compile(program, lnames)
        program.compose(len(self.expressions))

    cdef Expression _simplify(self, Context context):
        cdef list expressions = []
        cdef Expression expr
        for expr in self.expressions:
            expressions.append(expr._simplify(context))
        return sequence_pack(expressions)


cdef class Literal(Expression):
    cdef readonly Vector value

    def __init__(self, Vector value):
        self.value = value

    cdef void _compile(self, Program program, list lnames):
        program.literal(self.value.intern())

    cdef Expression _simplify(self, Context context):
        return self

    def __repr__(self):
        return f'Literal({self.value!r})'


cdef class Name(Expression):
    cdef readonly str name

    def __init__(self, str name):
        self.name = name

    cdef void _compile(self, Program program, list lnames):
        cdef int i, n=len(lnames)-1
        for i in range(len(lnames)):
            if self.name == <str>lnames[n-i]:
                program.local_load(i)
                break
        else:
            if self.name in dynamic_builtins:
                program.literal(dynamic_builtins[self.name])
            else:
                logger.warning("Name should have been removed by simplifier: {}", self.name)
                program.literal(null_)

    cdef Expression _simplify(self, Context context):
        if self.name in context.names:
            value = context.names[self.name]
            if value is None:
                return self
            elif isinstance(value, Function):
                return FunctionName(self.name)
            elif isinstance(value, Name):
                return (<Name>value)._simplify(context)
            else:
                return Literal(Vector.copy((<Vector>value)))
        elif (value := static_builtins.get(self.name)) is not None:
            return Literal(value)
        elif self.name not in dynamic_builtins:
            if context.captures is not None:
                context.captures.add(self.name)
            else:
                context.errors.add(f"Unbound name '{self.name}'")
                return NoOp
        return self

    def __repr__(self):
        return f'Name({self.name!r})'


cdef class FunctionName(Name):
    def __repr__(self):
        return f'FunctionName({self.name!r})'


cdef class Lookup(Expression):
    cdef readonly Expression key

    def __init__(self, Expression key):
        self.key = key

    cdef void _compile(self, Program program, list lnames):
        self.key._compile(program, lnames)
        program.lookup()

    cdef Expression _simplify(self, Context context):
        cdef Expression key = self.key._simplify(context)
        cdef Vector value
        if isinstance(key, Literal):
            if context.state is not None:
                value = context.state.get_item((<Literal>key).value)
                return Literal(value)
            return LookupLiteral((<Literal>key).value)
        return Lookup(key)

    def __repr__(self):
        return f'Lookup({self.key!r})'


cdef class LookupLiteral(Expression):
    cdef readonly Vector key

    def __init__(self, Vector key):
        self.key = key

    cdef void _compile(self, Program program, list lnames):
        program.lookup_literal(self.key.intern())

    cdef Expression _simplify(self, Context context):
        cdef Vector value
        if context.state is not None:
            value = context.state.get_item(self.key)
            return Literal(value)
        return LookupLiteral(self.key)

    def __repr__(self):
        return f'LookupLiteral({self.key!r})'


cdef class Range(Expression):
    cdef readonly Expression start
    cdef readonly Expression stop
    cdef readonly Expression step

    def __init__(self, Expression start, Expression stop, Expression step):
        self.start = start
        self.stop = stop
        self.step = step

    cdef void _compile(self, Program program, list lnames):
        self.start._compile(program, lnames)
        self.stop._compile(program, lnames)
        self.step._compile(program, lnames)
        program.range()

    cdef Expression _simplify(self, Context context):
        cdef Expression start = self.start._simplify(context)
        cdef Expression stop = self.stop._simplify(context)
        cdef Expression step = self.step._simplify(context)
        cdef Vector result
        if isinstance(start, Literal) and isinstance(stop, Literal) and isinstance(step, Literal):
            result = Vector.__new__(Vector)
            result.fill_range((<Literal>start).value, (<Literal>stop).value, (<Literal>step).value)
            return Literal(result)
        return Range(start, stop, step)

    def __repr__(self):
        return f'Range({self.start!r}, {self.stop!r}, {self.step!r})'


cdef class UnaryOperation(Expression):
    cdef readonly Expression expr

    def __init__(self, Expression expr):
        self.expr = expr

    cdef void _compile(self, Program program, list lnames):
        self.expr._compile(program, lnames)
        self._compile_op(program)

    cdef void _compile_op(self, Program program):
        raise NotImplementedError()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.expr!r})'


cdef class Negative(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.neg()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal((<Literal>expr).value.neg())
        if isinstance(expr, Negative):
            expr = Positive((<Negative>expr).expr)
            return expr._simplify(context)
        cdef MathsBinaryOperation maths
        if isinstance(expr, (Multiply, Divide)):
            maths = expr
            if isinstance(maths.left, Literal):
                expr = type(expr)(Negative(maths.left), maths.right)
                return expr._simplify(context)
            if isinstance(maths.right, Literal):
                expr = type(expr)(maths.left, Negative(maths.right))
                return expr._simplify(context)
        return Negative(expr)


cdef class Positive(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.pos()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal((<Literal>expr).value.pos())
        if isinstance(expr, (Negative, Positive, MathsBinaryOperation)):
            return expr._simplify(context)
        return Positive(expr)


cdef class Not(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.not_()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal(false_ if (<Literal>expr).value.as_bool() else true_)
        return Not(expr)


cdef class Ceil(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.ceil()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal((<Literal>expr).value.ceil())
        return Ceil(expr)


cdef class Floor(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.floor()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal((<Literal>expr).value.floor())
        return Floor(expr)


cdef class Fract(UnaryOperation):
    cdef void _compile_op(self, Program program):
        program.fract()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        if isinstance(expr, Literal):
            return Literal((<Literal>expr).value.fract())
        return Fract(expr)


cdef class BinaryOperation(Expression):
    cdef readonly Expression left
    cdef readonly Expression right

    def __init__(self, Expression left, Expression right):
        self.left = left
        self.right = right

    cdef void _compile(self, Program program, list lnames):
        self.left._compile(program, lnames)
        self.right._compile(program, lnames)
        self._compile_op(program)

    cdef void _compile_op(self, Program program):
        raise NotImplementedError()

    cdef Expression _simplify(self, Context context):
        cdef Expression left = self.left._simplify(context)
        cdef Expression right = self.right._simplify(context)
        cdef bint literal_left = isinstance(left, Literal)
        cdef bint literal_right = isinstance(right, Literal)
        if literal_left and literal_right:
            return Literal(self.op((<Literal>left).value, (<Literal>right).value))
        elif literal_left:
            if (expr := self.constant_left((<Literal>left).value, right)) is not None:
                return expr._simplify(context)
        elif literal_right:
            if (expr := self.constant_right(left, (<Literal>right).value)) is not None:
                return expr._simplify(context)
        return type(self)(left, right)

    cdef Vector op(self, Vector left, Vector right):
        raise NotImplementedError()

    cdef Expression constant_left(self, Vector left, Expression right):
        return None

    cdef Expression constant_right(self, Expression left, Vector right):
        return None

    def __repr__(self):
        return f'{self.__class__.__name__}({self.left!r}, {self.right!r})'


cdef class MathsBinaryOperation(BinaryOperation):
    cdef Expression _simplify(self, Context context):
        cdef Expression expr=BinaryOperation._simplify(self, context)
        if isinstance(expr, MathsBinaryOperation):
            if isinstance(expr.left, Positive):
                return (<Expression>type(expr)(expr.left.expr, expr.right))._simplify(context)
            elif isinstance(expr.right, Positive):
                return (<Expression>type(expr)(expr.left, expr.right.expr))._simplify(context)
        return expr


cdef class Add(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.add(right)

    cdef void _compile_op(self, Program program):
        program.add()

    cdef Expression constant_left(self, Vector left, Expression right):
        if left.eq(false_):
            return Positive(right)
        if isinstance(right, Negative):
            return Subtract(Literal(left), right.expr)

    cdef Expression constant_right(self, Expression left, Vector right):
        return self.constant_left(right, left)


cdef class Subtract(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.sub(right)

    cdef void _compile_op(self, Program program):
        program.sub()

    cdef Expression constant_left(self, Vector left, Expression right):
        if left.eq(false_):
            return Negative(right)

    cdef Expression constant_right(self, Expression left, Vector right):
        if right.eq(false_):
            return Positive(left)
        return Add(left, Literal(right.neg()))


cdef class Multiply(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.mul(right)

    cdef void _compile_op(self, Program program):
        program.mul()

    cdef Expression constant_left(self, Vector left, Expression right):
        if left.eq(true_):
            return Positive(right)
        if left.eq(minusone_):
            return Negative(right)
        cdef MathsBinaryOperation maths
        if isinstance(right, Add) or isinstance(right, Subtract):
            maths = right
            if isinstance(maths.left, Literal) or isinstance(maths.right, Literal):
                return type(maths)(Multiply(Literal(left), maths.left), Multiply(Literal(left), maths.right))
        elif isinstance(right, Multiply):
            maths = right
            if isinstance(maths.left, Literal):
                return Multiply(Multiply(Literal(left), maths.left), maths.right)
            if isinstance(maths.right, Literal):
                return Multiply(maths.left, Multiply(Literal(left), maths.right))
        elif isinstance(right, Divide):
            maths = right
            if isinstance(maths.left, Literal):
                return Divide(Multiply(Literal(left), maths.left), maths.right)
        elif isinstance(right, Negative):
            return Multiply(Literal(left.neg()), (<Negative>right).expr)

    cdef Expression constant_right(self, Expression left, Vector right):
        return self.constant_left(right, left)


cdef class Divide(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.truediv(right)

    cdef void _compile_op(self, Program program):
        program.truediv()

    cdef Expression constant_right(self, Expression left, Vector right):
        if right.eq(true_):
            return Positive(left)
        return Multiply(left, Literal(true_.truediv(right)))


cdef class FloorDivide(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.floordiv(right)

    cdef void _compile_op(self, Program program):
        program.floordiv()

    cdef Expression constant_right(self, Expression left, Vector right):
        if right.eq(true_):
            return Floor(left)


cdef class Modulo(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.mod(right)

    cdef void _compile_op(self, Program program):
        program.mod()

    cdef Expression constant_right(self, Expression left, Vector right):
        if right.eq(true_):
            return Fract(left)


cdef class Power(MathsBinaryOperation):
    cdef Vector op(self, Vector left, Vector right):
        return left.pow(right)

    cdef void _compile_op(self, Program program):
        program.pow()

    cdef Expression constant_right(self, Expression left, Vector right):
        if right.eq(true_):
            return Positive(left)


cdef class Comparison(BinaryOperation):
    pass


cdef class EqualTo(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.eq(right)

    cdef void _compile_op(self, Program program):
        program.eq()


cdef class NotEqualTo(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.ne(right)

    cdef void _compile_op(self, Program program):
        program.ne()


cdef class LessThan(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.lt(right)

    cdef void _compile_op(self, Program program):
        program.lt()


cdef class GreaterThan(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.gt(right)

    cdef void _compile_op(self, Program program):
        program.gt()


cdef class LessThanOrEqualTo(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.le(right)

    cdef void _compile_op(self, Program program):
        program.le()


cdef class GreaterThanOrEqualTo(Comparison):
    cdef Vector op(self, Vector left, Vector right):
        return left.ge(right)

    cdef void _compile_op(self, Program program):
        program.ge()


cdef class And(BinaryOperation):
    cdef void _compile(self, Program program, list lnames):
        end_label = program.new_label()
        self.left._compile(program, lnames)
        program.dup()
        program.branch_false(end_label)
        program.drop()
        self.right._compile(program, lnames)
        program.label(end_label)

    cdef Expression _simplify(self, Context context):
        cdef Expression left = self.left._simplify(context)
        if isinstance(left, Literal):
            if (<Literal>left).value.as_bool():
                return self.right._simplify(context)
            return left
        return And(left, self.right._simplify(context))


cdef class Or(BinaryOperation):
    cdef void _compile(self, Program program, list lnames):
        end_label = program.new_label()
        self.left._compile(program, lnames)
        program.dup()
        program.branch_true(end_label)
        program.drop()
        self.right._compile(program, lnames)
        program.label(end_label)

    cdef Expression _simplify(self, Context context):
        cdef Expression left = self.left._simplify(context)
        if isinstance(left, Literal):
            if not (<Literal>left).value.as_bool():
                return self.right._simplify(context)
            return left
        return Or(left, self.right._simplify(context))


cdef class Xor(BinaryOperation):
    cdef void _compile_op(self, Program program):
        program.xor()

    cdef Expression _simplify(self, Context context):
        cdef Expression left = self.left._simplify(context)
        cdef Expression right = self.right._simplify(context)
        cdef bint literal_left = isinstance(left, Literal)
        cdef bint literal_right = isinstance(right, Literal)
        if literal_left and not (<Literal>left).value.as_bool():
            return right
        if literal_right and not (<Literal>right).value.as_bool():
            return left
        if literal_left and literal_right:
            return Literal(false_)
        return Xor(left, right)


cdef class Slice(Expression):
    cdef readonly Expression expr
    cdef readonly Expression index

    def __init__(self, Expression expr, Expression index):
        self.expr = expr
        self.index = index

    cdef void _compile(self, Program program, list lnames):
        self.expr._compile(program, lnames)
        self.index._compile(program, lnames)
        program.slice()

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        cdef Expression index = self.index._simplify(context)
        cdef Vector expr_value
        cdef Vector index_value
        if isinstance(expr, Literal) and isinstance(index, Literal):
            expr_value = (<Literal>expr).value
            index_value = (<Literal>index).value
            return Literal(expr_value.slice(index_value))
        elif isinstance(index, Literal):
            index_value = (<Literal>index).value
            return FastSlice(expr, index_value)
        return Slice(expr, index)

    def __repr__(self):
        return f'Slice({self.expr!r}, {self.index!r})'


cdef class FastSlice(Expression):
    cdef readonly Expression expr
    cdef readonly Vector index

    def __init__(self, Expression expr, Vector index):
        self.expr = expr
        self.index = index

    cdef void _compile(self, Program program, list lnames):
        self.expr._compile(program, lnames)
        program.slice_literal(self.index.intern())

    cdef Expression _simplify(self, Context context):
        cdef Expression expr = self.expr._simplify(context)
        return FastSlice(expr, self.index)

    def __repr__(self):
        return f'FastSlice({self.expr!r}, {self.index!r})'


cdef class Call(Expression):
    cdef readonly Expression function
    cdef readonly tuple args
    cdef readonly tuple keyword_args

    def __init__(self, Expression function, tuple args, tuple keyword_args=None):
        self.function = function
        self.args = args
        self.keyword_args = keyword_args

    cdef void _compile(self, Program program, list lnames):
        cdef Expression expr
        cdef list names = []
        if self.args:
            for expr in self.args:
                expr._compile(program, lnames)
        cdef Binding keyword_arg
        if self.keyword_args:
            for keyword_arg in self.keyword_args:
                names.append(keyword_arg.name)
                keyword_arg.expr._compile(program, lnames)
        if not names and isinstance(self.function, Literal) \
                and (<Literal>self.function).value.length == 1 \
                and (<Literal>self.function).value.objects is not None \
                and not isinstance(function := (<Literal>self.function).value.objects[0], Function):
            program.call_fast(function, len(self.args) if self.args else 0)
        else:
            self.function._compile(program, lnames)
            program.call(len(self.args) if self.args else 0, tuple(names) if names else None)

    cdef Expression _simplify(self, Context context):
        cdef Expression function = self.function._simplify(context)
        cdef bint literal = isinstance(function, FunctionName) or \
            (isinstance(function, Literal) and (<Literal>function).value.objects is not None)
        cdef Expression arg, expr
        cdef list args = []
        if self.args:
            for arg in self.args:
                arg = arg._simplify(context)
                args.append(arg)
                if not isinstance(arg, Literal):
                    literal = False
        cdef list keyword_args = []
        cdef Binding binding
        if self.keyword_args:
            for binding in self.keyword_args:
                arg = binding.expr._simplify(context)
                keyword_args.append(Binding(binding.name, arg))
                if not isinstance(arg, Literal):
                    literal = False
        cdef list bindings
        cdef Function func_expr
        cdef dict kwargs
        cdef int i
        if isinstance(function, FunctionName):
            func_expr = context.names[(<FunctionName>function).name]
            assert not func_expr.captures, "Cannot inline functions with captured names"
            if not func_expr.recursive or literal:
                kwargs = {binding.name: binding.expr for binding in keyword_args}
                bindings = [PolyBinding(((<FunctionName>function).name,), func_expr)]
                for i, binding in enumerate(func_expr.parameters):
                    if i < len(args):
                        bindings.append(PolyBinding((binding.name,), <Expression>args[i]))
                    elif binding.name in kwargs:
                        bindings.append(PolyBinding((binding.name,), <Expression>kwargs[binding.name]))
                    elif binding.expr is not None:
                        bindings.append(PolyBinding((binding.name,), binding.expr))
                    else:
                        bindings.append(PolyBinding((binding.name,), Literal(null_)))
                expr = InlineLet(func_expr.expr, tuple(bindings))._simplify(context)
                return expr
        cdef list vector_args, results
        cdef Literal literal_arg
        if literal:
            vector_args = [literal_arg.value for literal_arg in args]
            kwargs = {binding.name: (<Literal>binding.expr).value for binding in keyword_args}
            results = []
            for func in (<Literal>function).value.objects:
                if callable(func):
                    try:
                        if hasattr(func, 'context_func'):
                            results.append(Literal(func(context, *vector_args, **kwargs)))
                        else:
                            results.append(Literal(func(*vector_args, **kwargs)))
                    except Exception:
                        break
            else:
                return sequence_pack(results)
        if isinstance(function, Literal) and len(args) == 1:
            if (<Literal>function).value == static_builtins['ceil']:
                return Ceil(args[0])
            elif (<Literal>function).value == static_builtins['floor']:
                return Floor(args[0])
            if (<Literal>function).value == static_builtins['fract']:
                return Fract(args[0])
        cdef Call call = Call(function, tuple(args), tuple(keyword_args))
        return call

    def __repr__(self):
        return f'Call({self.function!r}, {self.args!r}, {self.keyword_args!r})'


cdef class NodeModifier(Expression):
    cdef readonly Expression node


cdef class Tag(NodeModifier):
    cdef readonly str tag

    def __init__(self, Expression node, str tag):
        self.node = node
        self.tag = tag

    cdef void _compile(self, Program program, list lnames):
        self.node._compile(program, lnames)
        program.tag(self.tag)

    cdef Expression _simplify(self, Context context):
        cdef Expression node = self.node._simplify(context)
        cdef Vector nodes
        cdef list objects
        cdef Node n
        cdef int i
        if isinstance(node, Literal):
            nodes = (<Literal>node).value
            if nodes.isinstance(Node):
                objects = []
                for i in range(nodes.length):
                    n = nodes.objects[i].copy()
                    n.add_tag(self.tag)
                    objects.append(n)
                return Literal(Vector(objects))
        return Tag(node, self.tag)

    def __repr__(self):
        return f'Tag({self.node!r}, {self.tag!r})'


cdef class Attributes(NodeModifier):
    cdef readonly tuple bindings

    def __init__(self, Expression node, tuple bindings):
        self.node = node
        self.bindings = bindings

    cdef void _compile(self, Program program, list lnames):
        self.node._compile(program, lnames)
        cdef Binding binding
        for binding in self.bindings:
            binding.expr._compile(program, lnames)
            program.attribute(binding.name)

    cdef Expression _simplify(self, Context context):
        cdef Expression node = self
        cdef list bindings = []
        cdef Attributes attrs
        cdef Binding binding
        while isinstance(node, Attributes):
            attrs = <Attributes>node
            for binding in reversed(attrs.bindings):
                bindings.append(Binding(binding.name, binding.expr._simplify(context)))
            node = attrs.node
        node = node._simplify(context)
        cdef Vector nodes
        cdef int i
        cdef list objects
        if isinstance(node, Literal):
            nodes = (<Literal>node).value
            if nodes.isinstance(Node):
                objects = []
                for i in range(nodes.length):
                    objects.append(nodes.objects[i].copy())
                while bindings and isinstance((<Binding>bindings[-1]).expr, Literal):
                    binding = bindings.pop()
                    for i in range(nodes.length):
                        (<Node>objects[i]).set_attribute(binding.name, (<Literal>binding.expr).value)
                node = Literal(Vector(objects))
        if not bindings:
            return node
        bindings.reverse()
        return Attributes(node, tuple(bindings))

    def __repr__(self):
        return f'{self.__class__.__name__}({self.node!r}, {self.bindings!r})'


cdef class Append(NodeModifier):
    cdef readonly Expression children

    def __init__(self, Expression node, Expression children):
        self.node = node
        self.children = children

    cdef void _compile(self, Program program, list lnames):
        self.node._compile(program, lnames)
        self.children._compile(program, lnames)
        program.append()

    cdef Expression _simplify(self, Context context):
        cdef Expression node = self.node._simplify(context)
        cdef Expression children = self.children._simplify(context)
        cdef Vector nodes, childs
        cdef Node c, n
        cdef int i
        cdef list modifiers
        cdef Expression expr
        cdef NodeModifier modifier
        cdef list objects
        if isinstance(children, Literal):
            if isinstance(node, Literal):
                nodes = (<Literal>node).value
                childs = (<Literal>children).value
                if nodes.isinstance(Node) and childs.isinstance(Node):
                    objects = []
                    for i in range(nodes.length):
                        n = <Node>nodes.objects[i].copy()
                        for c in childs.objects:
                            n.append(c)
                        objects.append(n)
                    return Literal(Vector(objects))
            else:
                modifiers = []
                expr = node
                while isinstance(expr, NodeModifier):
                    modifiers.append(expr)
                    expr = (<NodeModifier>expr).node
                if isinstance(expr, Literal):
                    expr = Append(expr, children)._simplify(context)
                    while modifiers:
                        modifier = modifiers.pop()
                        if isinstance(modifier, Attributes):
                            expr = Attributes(expr, (<Attributes>modifier).bindings)
                        elif isinstance(modifier, Tag):
                            expr = Tag(expr, (<Tag>modifier).tag)
                        elif isinstance(modifier, Append):
                            expr = Append(expr, (<Append>modifier).children)
                    return expr
        elif isinstance(children, Sequence) and isinstance((<Sequence>children).expressions[0], Literal):
            node = Append(node, (<Sequence>children).expressions[0])
            children = Sequence((<Sequence>children).expressions[1:])
            return Append(node, children)._simplify(context)
        return Append(node, children)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.node!r}, {self.children!r})'


cdef class Binding:
    cdef readonly str name
    cdef readonly Expression expr

    def __init__(self, str name, Expression expr):
        self.name = name
        self.expr = expr

    def __repr__(self):
        return f'Binding({self.name!r}, {self.expr!r})'


cdef class PolyBinding:
    cdef readonly tuple names
    cdef readonly Expression expr

    def __init__(self, tuple names, Expression expr):
        self.names = names
        self.expr = expr

    def __repr__(self):
        return f'PolyBinding({self.names!r}, {self.expr!r})'


cdef class Let(Expression):
    cdef readonly tuple bindings

    def __init__(self, tuple bindings):
        self.bindings = bindings

    cdef void _compile(self, Program program, list lnames):
        cdef PolyBinding binding
        for binding in self.bindings:
            binding.expr._compile(program, lnames)
            program.local_push(len(binding.names))
            lnames.extend(binding.names)

    cdef Expression _simplify(self, Context context):
        cdef list remaining = []
        cdef PolyBinding binding
        cdef Expression expr
        cdef Vector value
        cdef str name
        cdef int i, n
        for binding in self.bindings:
            expr = binding.expr._simplify(context)
            if isinstance(expr, Literal):
                value = (<Literal>expr).value
                n = len(binding.names)
                if n == 1:
                    name = binding.names[0]
                    context.names[name] = value
                else:
                    for i, name in enumerate(binding.names):
                        context.names[name] = value.item(i)
            elif isinstance(expr, Name) and len(binding.names) == 1:
                name = binding.names[0]
                if (<Name>expr).name != name:
                    context.names[name] = expr
            else:
                for name in binding.names:
                    context.names[name] = None
                remaining.append(PolyBinding(binding.names, expr))
        if remaining:
            return Let(tuple(remaining))
        return NoOp

    def __repr__(self):
        return f'{self.__class__.__name__}({self.bindings!r})'


cdef class StoreGlobal(Expression):
    cdef readonly tuple bindings

    def __init__(self, tuple bindings):
        self.bindings = bindings

    cdef void _compile(self, Program program, list lnames):
        cdef Binding binding
        for binding in self.bindings:
            binding.expr._compile(program, lnames)
            program.store_global(binding.name)

    cdef Expression _simplify(self, Context context):
        return self

    def __repr__(self):
        return f'StoreGlobal({self.bindings!r})'


cdef class InlineLet(Expression):
    cdef readonly Expression body
    cdef readonly tuple bindings

    def __init__(self, Expression body, tuple bindings):
        self.body = body
        self.bindings = bindings

    cdef void _compile(self, Program program, list lnames):
        cdef PolyBinding binding
        cdef int n=len(lnames)
        for binding in self.bindings:
            binding.expr._compile(program, lnames)
            program.local_push(len(binding.names))
            lnames.extend(binding.names)
        self.body._compile(program, lnames)
        program.local_drop(len(self.bindings))
        while len(lnames) > n:
            lnames.pop()

    cdef Expression _simplify(self, Context context):
        cdef dict saved = context.names
        context.names = saved.copy()
        cdef list remaining = []
        cdef PolyBinding binding
        cdef Expression expr
        cdef Vector value
        cdef str name
        cdef int i, n
        for binding in self.bindings:
            expr = binding.expr._simplify(context)
            if isinstance(expr, Literal):
                value = (<Literal>expr).value
                n = len(binding.names)
                if n == 1:
                    name = binding.names[0]
                    context.names[name] = value
                else:
                    for i, name in enumerate(binding.names):
                        context.names[name] = value.item(i)
            elif isinstance(expr, Function) and len(binding.names) == 1:
                name = binding.names[0]
                context.names[name] = expr
            elif isinstance(expr, Name) and len(binding.names) == 1:
                name = binding.names[0]
                if (<Name>expr).name != name:
                    context.names[name] = expr
            else:
                for name in binding.names:
                    context.names[name] = None
                remaining.append(PolyBinding(binding.names, expr))
        body = self.body._simplify(context)
        context.names = saved
        if remaining:
            return InlineLet(body, tuple(remaining))
        return body

    def __repr__(self):
        return f'InlineLet({self.body!r}, {self.bindings!r})'


cdef class For(Expression):
    cdef readonly tuple names
    cdef readonly Expression source
    cdef readonly Expression body

    def __init__(self, tuple names, Expression source, Expression body):
        self.names = names
        self.source = source
        self.body = body

    cdef void _compile(self, Program program, list lnames):
        self.source._compile(program, lnames)
        program.begin_for()
        START = program.new_label()
        END = program.new_label()
        cdef int i, n=len(self.names)
        lnames.extend(self.names)
        program.literal(null_)
        program.local_push(n)
        program.label(START)
        program.next(n, END)
        self.body._compile(program, lnames)
        program.jump(START)
        program.label(END)
        program.local_drop(n)
        program.end_for_compose()
        for i in range(n):
            lnames.pop()

    cdef Expression _simplify(self, Context context):
        cdef Expression body, source=self.source._simplify(context)
        cdef list remaining = []
        cdef Vector values
        cdef dict saved = context.names
        context.names = saved.copy()
        cdef str name
        if not isinstance(source, Literal):
            for name in self.names:
                context.names[name] = None
            body = self.body._simplify(context)
            context.names = saved
            return For(self.names, source, body)
        values = (<Literal>source).value
        cdef int i=0, n=values.length
        while i < n:
            for name in self.names:
                context.names[name] = values.item(i)
                i += 1
            remaining.append(self.body._simplify(context))
        context.names = saved
        return sequence_pack(remaining)

    def __repr__(self):
        return f'For({self.names!r}, {self.source!r}, {self.body!r})'


cdef class IfCondition:
    cdef readonly Expression condition
    cdef readonly Expression then

    def __init__(self, Expression condition, Expression then):
        self.condition = condition
        self.then = then

    def __repr__(self):
        return f'IfCondition({self.condition!r}, {self.then!r})'


cdef class IfElse(Expression):
    cdef readonly tuple tests
    cdef readonly Expression else_

    def __init__(self, tuple tests, Expression else_):
        self.tests = tests
        self.else_ = else_

    cdef void _compile(self, Program program, list lnames):
        cdef IfCondition test
        END = program.new_label()
        for test in self.tests:
            NEXT = program.new_label()
            test.condition._compile(program, lnames)
            program.branch_false(NEXT)
            test.then._compile(program, lnames)
            program.jump(END)
            program.label(NEXT)
        if self.else_ is not None:
            self.else_._compile(program, lnames)
        else:
            program.literal(null_)
        program.label(END)

    cdef Expression _simplify(self, Context context):
        cdef list remaining = []
        cdef IfCondition test
        cdef Expression condition, then
        for test in self.tests:
            condition = test.condition._simplify(context)
            if isinstance(condition, Literal):
                if (<Literal>condition).value.as_bool():
                    then = test.then._simplify(context)
                    if not remaining:
                        return then
                    else:
                        return IfElse(tuple(remaining), then)
            else:
                then = test.then._simplify(context)
                remaining.append(IfCondition(condition, then))
        else_ = self.else_._simplify(context) if self.else_ is not None else None
        if remaining:
            return IfElse(tuple(remaining), else_)
        return NoOp if else_ is None else else_

    def __repr__(self):
        return f'IfElse({self.tests!r}, {self.else_!r})'


cdef class Function(Expression):
    cdef readonly str name
    cdef readonly tuple parameters
    cdef readonly Expression expr
    cdef readonly tuple captures
    cdef readonly bint recursive

    def __init__(self, str name, tuple parameters, Expression expr, tuple captures=None, bint recursive=False):
        self.name = name
        self.parameters = parameters
        self.expr = expr
        self.captures = captures
        self.recursive = recursive

    cdef void _compile(self, Program program, list lnames):
        cdef str name
        cdef list function_lnames = []
        assert self.captures is not None, "Simplifier must be run to correctly compile functions"
        cdef int i, n=len(lnames)-1
        for name in self.captures:
            for i in range(len(lnames)):
                if name == <str>lnames[n-i]:
                    program.local_load(i)
                    break
            else:
                logger.warning("Name should have been removed by simplifier: {}", name)
                program.literal(null_)
            function_lnames.append(name)
        function_lnames.append(self.name)
        cdef list parameters = []
        cdef Binding parameter
        for parameter in self.parameters:
            parameters.append(parameter.name)
            if parameter.expr is None:
                program.literal(null_)
            else:
                parameter.expr._compile(program, lnames)
            function_lnames.append(parameter.name)
        START = program.new_label()
        END = program.new_label()
        program.jump(END)
        program.label(START)
        self.expr._compile(program, function_lnames)
        program.exit()
        program.label(END)
        program.func(START, self.name, tuple(parameters), len(self.captures))
        program.local_push(1)
        lnames.append(self.name)

    cdef Expression _simplify(self, Context context):
        cdef list parameters = []
        cdef Binding parameter
        cdef Expression expr
        cdef bint literal = True
        for parameter in self.parameters:
            expr = parameter.expr._simplify(context) if parameter.expr is not None else None
            if expr is not None and not isinstance(expr, Literal):
                literal = False
            parameters.append(Binding(parameter.name, expr))
        cdef dict saved = context.names
        cdef str key
        cdef set captures = context.captures
        context.captures = set()
        context.names = {}
        for key, value in saved.items():
            if value is not None and key != self.name:
                context.names[key] = value
        for parameter in parameters:
            context.names[parameter.name] = None
        expr = self.expr._simplify(context)
        cdef recursive = self.name in context.captures
        if recursive:
            context.captures.discard(self.name)
        cdef Function function = Function(self.name, tuple(parameters), expr, tuple(context.captures), recursive)
        context.names = saved
        context.names[self.name] = function if literal and not context.captures else None
        if captures is not None:
            context.captures.difference_update(saved)
            captures.update(context.captures)
        context.captures = captures
        return function

    def __repr__(self):
        return f'Function({self.name!r}, {self.parameters!r}, {self.expr!r}, {self.captures!r}, {self.recursive!r})'



"""
Tests of the flitter language built-in functions
"""

import math
import unittest

from flitter.model import Vector, Context, StateDict, null
from flitter.language.functions import (uniform, normal, beta, counter, hypot, angle, split, ordv, chrv)


Tau = 2*math.pi


def all_isclose(xs, ys, rel_tol=1e-9, abs_tol=0):
    for x, y in zip(xs, ys):
        if not math.isclose(x, y, rel_tol=rel_tol, abs_tol=abs_tol):
            return False
    return True


class TestUniform(unittest.TestCase):
    FACTORY = uniform
    DISTRIBUTION = ('uniform',)
    LOWER = 0
    UPPER = 1
    P_VALUE = 0.05

    def test_creation(self):
        self.assertIsInstance(self.FACTORY(), self.FACTORY)
        self.assertEqual(hash(self.FACTORY()), hash(Vector()))
        self.assertEqual(hash(self.FACTORY(1.0)), hash(self.FACTORY(1.9)))

    def test_eq(self):
        source1 = self.FACTORY(1)
        source2 = self.FACTORY(1)
        self.assertIsNot(source1, source2)
        self.assertEqual(source1, source2)

    def test_null(self):
        self.assertEqual(self.FACTORY() + 1, null)
        self.assertEqual(1 + self.FACTORY(), null)
        self.assertEqual(self.FACTORY() * 1, null)
        self.assertEqual(self.FACTORY() / 1, null)
        self.assertEqual(self.FACTORY() % 1, null)
        self.assertEqual(self.FACTORY() // 1, null)
        self.assertEqual(self.FACTORY() ** 1, null)

    def test_indexing_and_overlap(self):
        source = self.FACTORY()
        last_xs = None
        for i in range(-100, 100):
            source_i = source[i]
            self.assertIsInstance(source_i, float)
            with self.subTest(i=i):
                xs = source[i:i+10]
                self.assertIsInstance(xs, Vector)
                self.assertEqual(len(xs), 10)
                self.assertEqual(xs[0], source_i)
                for x in xs:
                    self.assertIsInstance(x, float)
                if last_xs is not None:
                    self.assertNotEqual(last_xs, xs)
                    self.assertEqual(last_xs[1:10], xs[0:9])
                last_xs = xs

    def test_reproducability(self):
        source1 = self.FACTORY(1.0)
        source2 = self.FACTORY(1.1)
        self.assertIsNot(source1, source2)
        self.assertEqual(source1[:10_000], source2[:10_000])

    def test_distribution(self):
        from scipy.stats import kstest
        for i in range(2):
            with self.subTest(i=i):
                result = kstest(self.FACTORY(i)[:10_000_000], *self.DISTRIBUTION)
                self.assertGreater(result.pvalue, self.P_VALUE)

    def test_range(self):
        source = self.FACTORY()
        for i in range(10000):
            with self.subTest(i=i):
                x = source[i]
                self.assertTrue(x >= self.LOWER)
                self.assertTrue(x < self.UPPER)

    def test_apparent_entropy(self):
        from struct import pack
        from zlib import compress
        data = pack('<1000L', *(int(n * (1 << 32)) for n in self.FACTORY()[:1000]))
        compressed = compress(data, 9)
        self.assertGreater(len(compressed) / len(data), 0.99)


class TestBeta(TestUniform):
    FACTORY = beta
    DISTRIBUTION = ('beta', (2, 2))


class TestNormal(TestUniform):
    FACTORY = normal
    DISTRIBUTION = ('norm',)
    LOWER = -10
    UPPER = 10

    def test_apparent_entropy(self):
        pass


class TestCounter(unittest.TestCase):
    def setUp(self):
        self.state = StateDict()
        self.context = Context(state=self.state)
        self.counter_id = Vector('counter')

    def test_simple(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock), count)
            clock += 0.5
            count += 0.5

    def test_offset(self):
        clock = Vector(10)
        count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock), count)
            clock += 0.5
            count += 0.5

    def test_fast(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(2)), count)
            clock += 0.5
            count += 1.0

    def test_slow(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(0.5)), count)
            clock += 0.5
            count += 0.25

    def test_speed_up(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(0.5)), count)
            clock += 0.5
            count += 0.25
        while clock < 10:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(2)), count)
            clock += 0.5
            count += 1.0

    def test_slow_down(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(2)), count)
            clock += 0.5
            count += 1.0
        while clock < 10:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(0.5)), count)
            clock += 0.5
            count += 0.25

    def test_pause(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock), count)
            clock += 0.5
            count += 0.5
        while clock < 10:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(0)), count)
            clock += 0.5

    def test_reverse(self):
        clock = count = Vector(0)
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock), count)
            clock += 0.5
            count += 0.5
        while clock < 10:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(-2)), count)
            clock += 0.5
            count -= 1.0

    def test_vector_speed(self):
        clock = Vector(0)
        count = Vector([0, 0])
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector([1, 2])), count)
            clock += 0.5
            count += [0.5, 1]

    def test_vector_clock(self):
        clock = Vector([0, 0])
        count = Vector([0, 0])
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector(2)), count)
            clock += [0.5, 1]
            count += [1, 2]

    def test_vector_clock_and_speed(self):
        clock = Vector([0, 0])
        count = Vector([0, 0])
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector([1, 2])), count)
            clock += [1, 0.5]
            count += [1, 1]

    def test_state_changes(self):
        clock = Vector([0, 0])
        count = counter(self.context, self.counter_id, clock, Vector([1, 2]))
        self.assertTrue(self.counter_id in self.state.changed_keys)
        self.state.clear_changed()
        while clock < 5:
            self.assertEqual(counter(self.context, self.counter_id, clock, Vector([1, 2])), count)
            self.assertFalse(self.counter_id in self.state.changed_keys)
            clock += [1, 0.5]
            count += [1, 1]
        self.assertEqual(counter(self.context, self.counter_id, clock, Vector([1, -2])), count)
        self.assertTrue(self.counter_id in self.state.changed_keys)

    def test_read_without_update(self):
        clock = count = Vector(0)
        count = counter(self.context, self.counter_id, clock, Vector(2))
        self.assertTrue(self.counter_id in self.state.changed_keys)
        self.state.clear_changed()
        clock += 0.5
        count += 1.0
        read_count = counter(self.context, self.counter_id, clock)
        self.assertFalse(self.counter_id in self.state.changed_keys)
        self.assertEqual(count, read_count)


class TestTrig(unittest.TestCase):
    def setUp(self):
        self.a = Vector([1, 2, 3, 4])
        self.b = Vector([4, 5, 6, 7])
        self.c = Vector([2])

    def test_hypot_one_arg(self):
        self.assertEqual(hypot(), null)
        self.assertEqual(hypot(null), null)
        self.assertTrue(all_isclose(hypot(self.a), Vector(math.sqrt(30))))
        self.assertTrue(all_isclose(hypot(self.c), self.c))

    def test_hypot_multiple_args(self):
        self.assertEqual(hypot(self.a, null), null)
        self.assertTrue(all_isclose(hypot(self.a, self.b), Vector([math.sqrt(17), math.sqrt(29), math.sqrt(45), math.sqrt(65)])))
        self.assertTrue(all_isclose(hypot(self.a, self.c), Vector([math.sqrt(5), math.sqrt(8), math.sqrt(13), math.sqrt(20)])))
        self.assertTrue(all_isclose(hypot(self.a, self.b, self.c),
                                    Vector([math.sqrt(21), math.sqrt(33), math.sqrt(49), math.sqrt(69)])))

    def test_angle_one_arg(self):
        self.assertEqual(angle(null), null)
        self.assertTrue(all_isclose(angle(self.a), Vector([math.atan2(2, 1)/Tau, math.atan2(4, 3)/Tau])))

    def test_angle_multiple_args(self):
        self.assertEqual(angle(self.a, null), null)
        self.assertEqual(angle(null, self.a), null)
        self.assertTrue(all_isclose(angle(self.a, self.b),
                                    Vector([math.atan2(4, 1)/Tau, math.atan2(5, 2)/Tau, math.atan2(6, 3)/Tau, math.atan2(7, 4)/Tau])))
        self.assertTrue(all_isclose(angle(self.a, self.c),
                                    Vector([math.atan2(2, 1)/Tau, math.atan2(2, 2)/Tau, math.atan2(2, 3)/Tau, math.atan2(2, 4)/Tau])))


class TestStringFuncs(unittest.TestCase):
    def test_ord(self):
        self.assertEqual(ordv(null), null)
        self.assertEqual(ordv(Vector('A')), Vector([65]))
        self.assertEqual(ordv(Vector('AB')), Vector([65, 66]))
        self.assertEqual(ordv(Vector(['A', 'B'])), Vector([65, 66]))

    def test_chr(self):
        self.assertEqual(chrv(null), null)
        self.assertEqual(chrv(Vector('A')), null)
        self.assertEqual(chrv(Vector([65])), Vector(['A']))
        self.assertEqual(chrv(Vector([65, 66])), Vector(['AB']))

    def test_split(self):
        self.assertEqual(split(null), null)
        self.assertEqual(split(Vector(['Hello world!'])), Vector(['Hello world!']))
        self.assertEqual(split(Vector(['Hello world!\n'])), Vector(['Hello world!']))
        self.assertEqual(split(Vector(['Hello\nworld!'])), Vector(['Hello', 'world!']))
        self.assertEqual(split(Vector(['Hello\nworld!\n'])), Vector(['Hello', 'world!']))
        self.assertEqual(split(Vector(['Hello\n\nworld!\n'])), Vector(['Hello', '', 'world!']))



"""
The main Flitter engine
"""

import asyncio
import gc
from pathlib import Path
import pickle

from loguru import logger

from ..cache import SharedCache
from ..clock import BeatCounter, system_clock
from ..language.vm import log_vm_stats
from ..model import Vector, StateDict, Context, null, numbers_cache_counts, empty_numbers_cache
from ..render import get_renderer


class EngineController:
    def __init__(self, target_fps=60, screen=0, fullscreen=False, vsync=False, state_file=None,
                 autoreset=None, state_simplify_wait=0, realtime=True, defined_names=None, vm_stats=False,
                 run_time=None, offscreen=False, window_gamma=1):
        self.default_fps = target_fps
        self.target_fps = target_fps
        self.realtime = realtime
        self.screen = screen
        self.fullscreen = fullscreen
        self.vsync = vsync
        self.offscreen = offscreen
        self.window_gamma = window_gamma
        self.autoreset = autoreset
        self.state_simplify_wait = state_simplify_wait
        if defined_names:
            self.defined_names = {key: Vector.coerce(value) for key, value in defined_names.items()}
        else:
            self.defined_names = {}
        self.vm_stats = vm_stats
        self.run_time = run_time
        self.state_file = Path(state_file) if state_file is not None else None
        if self.state_file is not None and self.state_file.exists():
            logger.info("Recover state from state file: {}", self.state_file)
            with open(self.state_file, 'rb') as file:
                self.global_state = pickle.load(file)
        else:
            self.global_state = {}
        self.global_state_dirty = False
        self.state = None
        self.state_timestamp = None
        self.renderers = {}
        self.counter = BeatCounter()
        self.pages = []
        self.switch_page = None
        self.current_page = None
        self.current_path = None
        self._references = {}

    def load_page(self, filename):
        page_number = len(self.pages)
        path = SharedCache.get_with_root(filename, '.')
        self.pages.append((path, self.global_state.setdefault(str(path), StateDict())))
        logger.info("Added code page {}: {}", page_number, path)
        if page_number == 0:
            self.switch_to_page(0)

    def switch_to_page(self, page_number):
        if self.pages is not None and 0 <= page_number < len(self.pages):
            path, state = self.pages[page_number]
            self.state = state
            self.state_timestamp = system_clock()
            self.current_path = path
            self.current_page = page_number
            SharedCache.set_root(self.current_path)
            logger.info("Switched to page {}: {}", page_number, self.current_path)
            if counter_state := self.state['_counter']:
                tempo, quantum, start = counter_state
                self.counter.reset(tempo, int(quantum), start)
                logger.info("Restore counter at beat {:.1f}, tempo {:.1f}, quantum {}", self.counter.beat, self.counter.tempo, self.counter.quantum)
            else:
                self.counter.reset()
            for renderers in self.renderers.values():
                for renderer in renderers:
                    renderer.purge()

    def has_next_page(self):
        return self.current_page < len(self.pages) - 1

    def next_page(self):
        if self.current_page < len(self.pages) - 1:
            self.switch_page = self.current_page + 1

    def has_previous_page(self):
        return self.current_page > 0

    def previous_page(self):
        if self.current_page > 0:
            self.switch_page = self.current_page - 1

    async def update_renderers(self, root, **kwargs):
        nodes_by_kind = {}
        for node in root.children:
            nodes_by_kind.setdefault(node.kind, []).append(node)
        tasks = []
        references = dict(self._references)
        for kind, nodes in nodes_by_kind.items():
            renderer_class = get_renderer(kind)
            if renderer_class is not None:
                renderers = self.renderers.setdefault(kind, [])
                count = 0
                for node in nodes:
                    if count == len(renderers):
                        renderer = renderer_class(**kwargs)
                        renderers.append(renderer)
                    tasks.append(asyncio.create_task(renderers[count].update(self, node, references=references, **kwargs)))
                    count += 1
                while len(renderers) > count:
                    renderers.pop().destroy()
        for kind in list(self.renderers):
            if kind not in nodes_by_kind:
                for renderer in self.renderers.pop(kind):
                    renderer.destroy()
        try:
            await asyncio.gather(*tasks)
        except Exception:
            for task in tasks:
                if not task.done():
                    task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise
        return references

    def handle_pragmas(self, pragmas, timestamp):
        tempo = pragmas.get('tempo', null).match(1, float, 120)
        if tempo != self.counter.tempo:
            self.counter.set_tempo(tempo, timestamp)
        quantum = pragmas.get('quantum', null).match(1, float, 4)
        if quantum != self.counter.quantum:
            self.counter.set_quantum(quantum, timestamp)
        self.target_fps = pragmas.get('fps', null).match(1, float, self.default_fps)

    def reset_state(self):
        self.state.clear()
        self.state_timestamp = None
        self.global_state_dirty = True

    async def run(self):
        try:
            frames = []
            start_time = frame_time = system_clock() if self.realtime else self.counter.start
            last = self.counter.beat_at_time(frame_time)
            dump_time = frame_time
            execution = render = housekeeping = 0
            performance = 1
            run_program = current_program = errors = logs = None
            gc.disable()
            while self.run_time is None or int(round((frame_time - start_time) * self.target_fps)) < int(round(self.run_time * self.target_fps)):
                housekeeping -= system_clock()

                beat = self.counter.beat_at_time(frame_time)
                delta = beat - last
                last = beat
                names = {'beat': beat, 'quantum': self.counter.quantum, 'tempo': self.counter.tempo,
                         'delta': delta, 'clock': frame_time, 'performance': performance,
                         'fps': self.target_fps, 'realtime': self.realtime, 'window_gamma': self.window_gamma,
                         'screen': self.screen, 'fullscreen': self.fullscreen, 'vsync': self.vsync, 'offscreen': self.offscreen}

                program = self.current_path.read_flitter_program(static=self.defined_names, dynamic=names)
                if program is not current_program:
                    level = 'SUCCESS' if current_program is None else 'INFO'
                    logger.log(level, "Loaded page {}: {}", self.current_page, self.current_path)
                    run_program = current_program = program
                    errors = set()
                    logs = set()

                if current_program is not None and run_program is current_program and self.state_simplify_wait and self.state_timestamp is not None and \
                        system_clock() > self.state_timestamp + self.state_simplify_wait:
                    simplify_time = -system_clock()
                    top = current_program.top.simplify(state=self.state, dynamic=names)
                    now = system_clock()
                    simplify_time += now
                    compile_time = -now
                    run_program = top.compile(initial_lnames=tuple(names))
                    run_program.set_path(current_program.path)
                    run_program.set_top(top)
                    compile_time += system_clock()
                    logger.debug("Simplified on state to {} instructions in -/{:.1f}/{:.1f}ms",
                                 len(run_program), simplify_time*1000, compile_time*1000)

                if self.state.changed:
                    self.global_state_dirty = True
                    self.state_timestamp = system_clock()
                    self.state.clear_changed()
                    if run_program is not current_program:
                        logger.debug("Undo simplification on state")
                        run_program = current_program

                now = system_clock()
                housekeeping += now
                execution -= now
                if run_program is not None:
                    context = Context(names={key: Vector.coerce(value) for key, value in names.items()},
                                      state=self.state, references=self._references)
                    run_program.run(context, record_stats=self.vm_stats)
                else:
                    context = Context()
                self.handle_pragmas(context.pragmas, frame_time)

                new_errors = context.errors.difference(errors) if errors is not None else context.errors
                errors = context.errors
                for error in new_errors:
                    logger.error("Execution error: {}", error)
                new_logs = context.logs.difference(logs) if logs is not None else context.logs
                logs = context.logs
                for log in new_logs:
                    print(log)
                now = system_clock()
                execution += now
                render -= now

                self._references = await self.update_renderers(context.root, **names)

                now = system_clock()
                render += now
                housekeeping -= now

                del context
                SharedCache.clean()

                self.state['_counter'] = self.counter.tempo, self.counter.quantum, self.counter.start

                if self.autoreset and self.state_timestamp is not None and system_clock() > self.state_timestamp + self.autoreset:
                    logger.debug("Auto-reset state")
                    self.reset_state()
                    current_program = program

                if self.global_state_dirty and self.state_file is not None and frame_time > dump_time + 1:
                    logger.debug("Saving state")
                    with open(self.state_file, 'wb') as file:
                        pickle.dump(self.global_state, file)
                    self.global_state_dirty = False
                    dump_time = frame_time

                if self.switch_page is not None:
                    if self.autoreset:
                        self.reset_state()
                    self.switch_to_page(self.switch_page)
                    self.switch_page = None
                    run_program = current_program = None
                    performance = 1
                    count = gc.collect(2)
                    logger.trace("Collected {} objects (full collection)", count)

                elif count := gc.collect(0):
                    logger.trace("Collected {} objects", count)

                now = system_clock()
                frames.append(now)
                frame_period = now - frame_time
                housekeeping += now
                frame_time += 1 / self.target_fps
                if self.realtime:
                    wait_time = frame_time - now
                    performance = min(performance + 0.001, 2) if wait_time > 0.001 else max(0.5, performance - 0.01)
                else:
                    wait_time = 0.001
                    performance = 1
                if wait_time > 0:
                    await asyncio.sleep(wait_time)
                else:
                    logger.trace("Slow frame - {:.0f}ms", frame_period * 1000)
                    await asyncio.sleep(0)
                    frame_time = system_clock()

                if len(frames) > 1 and frames[-1] - frames[0] > 5:
                    nframes = len(frames) - 1
                    fps = nframes / (frames[-1] - frames[0])
                    logger.info("{:4.1f}fps; {:4.1f}/{:4.1f}/{:4.1f}ms (run/render/sys); perf {:.2f}",
                                fps, 1000 * execution / nframes, 1000 * render / nframes, 1000 * housekeeping / nframes, performance)
                    frames = frames[-1:]
                    execution = render = housekeeping = 0
                    logger.trace("State dictionary size: {} keys", len(self.state))
                    if run_program is not None:
                        logger.trace("VM stack size: {:d}", run_program.stack.size)

        finally:
            self.global_state = {}
            self._references = {}
            self.pages = []
            program = run_program = current_program = context = None
            SharedCache.clean(0)
            for renderers in self.renderers.values():
                while renderers:
                    renderers.pop().destroy()
            if self.vm_stats:
                log_vm_stats()
            count = gc.collect(2)
            logger.trace("Collected {} objects (full collection)", count)
            counts = numbers_cache_counts()
            if counts:
                logger.debug("Numbers cache: {}", ", ".join(f"{size}x{count}" for size, count in counts.items()))
            empty_numbers_cache()
            gc.enable()



"""
Ableton Push 2 API

As per official hardware documentation at: https://github.com/Ableton/push-interface
"""

import asyncio
from contextlib import asynccontextmanager

from loguru import logger
import numpy as np
import rtmidi2
import skia
import usb.core

from ..clock import BeatCounter, system_clock
from .constants import (MIDI, Command, Animation, Control, Note, Encoder, TouchStripFlags,
                        BUTTONS, COLOR_BUTTONS, ENCODER_CONTROLS)
from .events import (PadPressed, PadHeld, PadReleased,
                     ButtonPressed, ButtonReleased,
                     EncoderTouched, EncoderTurned, EncoderReleased,
                     TouchStripTouched, TouchStripDragged, TouchStripReleased)
from .palette import SimplePalette


class Push2CommunicationError(Exception):
    pass


class Push2:
    USB_VENDOR = 0x2982
    USB_PRODUCT = 0x1967
    MIDI_PORT_NAME = 'Ableton Push 2 User Port'
    SCREEN_HEIGHT = 160
    SCREEN_WIDTH = 960
    SCREEN_STRIDE = 1024
    SYSEX_ID = [0x00, 0x21, 0x1D, 0x01, 0x01]
    FRAME_HEADER = bytes([0xFF, 0xCC, 0xAA, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00])

    @classmethod
    def test_presence(cls):
        return usb.core.find(idVendor=cls.USB_VENDOR, idProduct=cls.USB_PRODUCT) is not None

    def __init__(self, palette=None):
        self._palette = palette if palette is not None else SimplePalette()
        self._screen_data = bytearray(self.SCREEN_HEIGHT * self.SCREEN_STRIDE * 2)
        self._screen_array = np.ndarray(buffer=self._screen_data, shape=(self.SCREEN_HEIGHT, self.SCREEN_STRIDE), dtype='uint16')
        self._surface_array = np.zeros((self.SCREEN_HEIGHT, self.SCREEN_WIDTH, 4), dtype='uint8')
        self._screen_surface = skia.Surface(self._surface_array)
        self._screen_update = asyncio.Condition()
        self._loop = None
        self._usb_device = None
        self._screen_endpoint = None
        self._midi_out = None
        self._midi_in = None
        self._receive_queue = asyncio.Queue()
        self._last_receive_timestamp = None
        self._counter = None
        self._clock_task = None
        self._screen_task = None

    @property
    def counter(self):
        return self._counter

    async def start(self, counter=None):
        self._counter = BeatCounter() if counter is None else counter
        self._loop = asyncio.get_running_loop()
        self._usb_device = usb.core.find(idVendor=self.USB_VENDOR, idProduct=self.USB_PRODUCT)
        if self._usb_device is None:
            raise RuntimeError("Cannot locate USB device")
        self._usb_device.reset()
        self._usb_device.set_configuration()
        self._screen_endpoint = self._usb_device.get_active_configuration()[0, 0][0]
        self._midi_out = rtmidi2.MidiOut()
        for i, name in enumerate(rtmidi2.get_out_ports()):
            if name == self.MIDI_PORT_NAME:
                self._midi_out.open_port(i)
                break
        else:
            raise RuntimeError("Cannot open MIDI output port: " + self.MIDI_PORT_NAME)
        self._midi_in = rtmidi2.MidiIn()
        for i, name in enumerate(rtmidi2.get_in_ports()):
            if name == self.MIDI_PORT_NAME:
                self._midi_in.open_port(i)
                break
        else:
            raise RuntimeError("Cannot open MIDI input port: " + self.MIDI_PORT_NAME)
        self._midi_in.callback = self._receive_callback
        self._send_midi([MIDI.STOP])
        self._send_sysex(Command.SET_MIDI_MODE, 1)
        self._send_sysex(Command.SET_AFTERTOUCH_MODE, 1)
        self._send_sysex(Command.SET_TOUCHSTRIP_CONFIG, TouchStripFlags.HOST_CONTROL | TouchStripFlags.LEDS_POINT)
        self.apply_palette()
        self._clock_task = asyncio.create_task(self._run_clock())
        self._screen_task = asyncio.create_task(self._run_screen())

    @property
    def palette(self):
        return self._palette

    @palette.setter
    def palette(self, palette):
        self._palette = palette
        self.apply_palette()

    def apply_palette(self):
        for i in range(128):
            r, g, b = self._palette.index_to_rgb_led(i)
            w = self._palette.index_to_white_led(i)
            self._send_sysex(Command.SET_COLOR_PALETTE_ENTRY, i, r & 0x7f, r >> 7, g & 0x7f, g >> 7, b & 0x7f, b >> 7, w & 0x7f, w >> 7)
        self._send_sysex(Command.REAPPLY_COLOR_PALETTE)

    async def stop(self):
        if self._clock_task is not None:
            self._clock_task.cancel()
            await self._clock_task
            self._clock_task = None
        if self._midi_out is not None:
            for number in BUTTONS:
                self.set_button_white(number, 0)
            for number in range(64):
                self.set_pad_rgb(number, 0, 0, 0)
            self.set_touch_strip_position(0)
            self._midi_out.close_port()
            self._midi_out = None
        if self._midi_in is not None:
            self._midi_in.close_port()
            self._midi_in = None
        if self._screen_task is not None:
            self._screen_task.cancel()
            await self._screen_task
            self._screen_task = None
        if self._usb_device is not None:
            self._screen_endpoint = None
            self._usb_device = None
        self._receive_queue = asyncio.Queue()
        self._last_receive_timestamp = None
        self._counter = None
        self._loop = None

    def _send_midi(self, message):
        logger.trace("Send MIDI - {}", " ".join(f"{b:02x}" for b in message))
        self._midi_out.send_raw(*message)

    def _receive_callback(self, message, delta):
        now = system_clock()
        timestamp = now if self._last_receive_timestamp is None else min(now, self._last_receive_timestamp + delta)
        self._last_receive_timestamp = timestamp
        logger.trace("Received @ {:.2f} MIDI - {}", timestamp, " ".join(f"{b:02x}" for b in message))
        self._loop.call_soon_threadsafe(self._receive_queue.put_nowait, (message, timestamp))

    def _send_sysex(self, cmd: Command, *args):
        self._send_midi([MIDI.START_OF_SYSEX] + self.SYSEX_ID + [cmd] + list(args) + [MIDI.END_OF_SYSEX])

    def set_pad_rgb(self, number, r, g, b, animation=Animation.NONE):
        self.set_pad_index(number, self._palette.rgb_to_index(r, g, b), animation)

    def set_pad_index(self, number, index, animation=Animation.NONE):
        assert number < 64
        assert index in range(0, 128)
        self._send_midi([MIDI.NOTE_ON + animation, number + Note.PAD_0_0, index])

    def set_button_rgb(self, number, r, g, b, animation=Animation.NONE):
        if number in COLOR_BUTTONS:
            self.set_button_index(number, self._palette.rgb_to_index(r, g, b), animation)
        else:
            w = 0.299*r + 0.587*g + 0.114*b
            self.set_button_index(number, self._palette.white_to_index(w), animation)

    def set_button_white(self, number, w, animation=Animation.NONE):
        if number in COLOR_BUTTONS:
            self.set_button_index(number, self._palette.rgb_to_index(w, w, w), animation)
        else:
            self.set_button_index(number, self._palette.white_to_index(w), animation)

    def set_button_index(self, number, index, animation=Animation.NONE):
        assert number in BUTTONS
        assert index in range(0, 128)
        self._send_midi([MIDI.CONTROL_CHANGE + animation, number, index])

    def set_touch_strip_position(self, position):
        value = int(min(max(0, position), 1) * 255) << 6
        self._send_midi([MIDI.PITCH_BEND_CHANGE, value & 0x7f, value >> 7])

    def set_led_brightness(self, brightness):
        value = int(round(min(max(0, brightness), 1) * 127))
        self._send_sysex(Command.SET_LED_BRIGHTNESS, value)

    def set_display_brightness(self, brightness):
        value = int(round(min(max(0, brightness), 1) * 255))
        self._send_sysex(Command.SET_DISPLAY_BRIGHTNESS, value & 0x7f, value >> 7)

    def get_event(self):
        try:
            message, timestamp = self._receive_queue.get_nowait()
        except asyncio.QueueEmpty:
            return None
        return self.process_message(message, timestamp)

    async def wait_event(self):
        if self._clock_task.done():
            await self._clock_task
        if self._screen_task.done():
            await self._screen_task
        message, timestamp = await self._receive_queue.get()
        return self.process_message(message, timestamp)

    def process_message(self, message, timestamp):
        if message[0] == MIDI.NOTE_ON:
            if Note.PAD_0_0 <= message[1] <= Note.PAD_7_7:
                number = message[1] - Note.PAD_0_0
                column, row = number % 8, number // 8
                return PadPressed(timestamp=timestamp, number=number, column=column, row=row, pressure=message[2])
            if message[1] in set(Encoder):
                if message[2] == 0x7f:
                    return EncoderTouched(timestamp=timestamp, number=Encoder(message[1]))
                if message[2] == 0x00:
                    return EncoderReleased(timestamp=timestamp, number=Encoder(message[1]))
            if message[1] == Note.TOUCH_STRIP:
                if message[2] == 0x7f:
                    return TouchStripTouched(timestamp=timestamp)
                if message[2] == 0x00:
                    return TouchStripReleased(timestamp=timestamp)
        elif message[0] == MIDI.POLYPHONIC_PRESSURE:
            if Note.PAD_0_0 <= message[1] <= Note.PAD_7_7:
                number = message[1] - Note.PAD_0_0
                column, row = number % 8, number // 8
                return PadHeld(timestamp=timestamp, number=number, column=column, row=row, pressure=message[2])
        elif message[0] == MIDI.NOTE_OFF:
            if Note.PAD_0_0 <= message[1] <= Note.PAD_7_7:
                number = message[1] - Note.PAD_0_0
                column, row = number % 8, number // 8
                return PadReleased(timestamp=timestamp, number=number, column=column, row=row)
        elif message[0] == MIDI.CONTROL_CHANGE:
            if message[1] in BUTTONS:
                if message[2] == 0x7f:
                    return ButtonPressed(timestamp=timestamp, number=Control(message[1]))
                if message[2] == 0x00:
                    return ButtonReleased(timestamp=timestamp, number=Control(message[1]))
            if message[1] in ENCODER_CONTROLS:
                amount = message[2] if message[2] < 64 else message[2] - 128
                return EncoderTurned(timestamp=timestamp, number=ENCODER_CONTROLS[message[1]], amount=amount)
        elif message[0] == MIDI.PITCH_BEND_CHANGE:
            position = (message[2] << 7) + message[1]
            return TouchStripDragged(timestamp=timestamp, position=position)
        logger.warning("Unrecognised MIDI - {}", " ".join(f"{b:02x}" for b in message))

    @asynccontextmanager
    async def screen_context(self):
        if self._screen_task is None:
            raise TypeError("Screen task not running")
        async with self._screen_update:
            with self._screen_surface as ctx:
                ctx.save()
                yield ctx
                ctx.restore()
            self._screen_update.notify()

    async def _run_clock(self):
        logger.trace("Started Push2 clock task")
        try:
            tick = int(self._counter.beat * 24)
            while True:
                if tick % (self._counter.quantum * 24) == 0:
                    self._send_midi([MIDI.START])
                else:
                    self._send_midi([MIDI.CLOCK])
                tick += 1
                await self._counter.wait_for_beat(tick / 24)
        except asyncio.CancelledError:
            logger.trace("Stopped Push2 clock task")
            self._send_midi([MIDI.STOP])
        except Exception:
            logger.exception("Unexpected exception in Push2 clock task")

    async def _run_screen(self):
        logger.trace("Started Push2 screen task")
        try:
            async with self._screen_update:
                while True:
                    await asyncio.to_thread(self._update_screen)
                    try:
                        await asyncio.wait_for(self._screen_update.wait(), timeout=1)
                    except asyncio.TimeoutError:
                        pass
        except asyncio.CancelledError:
            logger.trace("Stopped Push2 screen task")
            self._surface_array[:, :, :] = 0
            self._update_screen()
        except usb.core.USBError as exc:
            raise Push2CommunicationError("Error sending screen data") from exc
        except Exception:
            logger.exception("Unexpected exception in Push2 screen task")
            raise

    def _update_screen(self):
        self._screen_array[:, :] = 0
        pixels = self._screen_array[:, :self._surface_array.shape[1]]
        pixels += self._surface_array[:, :, 2] >> 3
        pixels <<= 6
        pixels += self._surface_array[:, :, 1] >> 2
        pixels <<= 5
        pixels += self._surface_array[:, :, 0] >> 3
        self._screen_array[:, 0::2] ^= 0xF3E7
        self._screen_array[:, 1::2] ^= 0xFFE7
        self._screen_endpoint.write(self.FRAME_HEADER)
        self._screen_endpoint.write(self._screen_data)



"""
Clock API

Provides a monotonically(*) increasing beat counter with a given tempo.

(*) The beat counter can actually go backwards by up to `backslip_limit` beat(s)
    when adjusting the phase with `set_phase`.
"""

import asyncio
import time

from numpy.polynomial.polynomial import polyfit


system_clock = time.perf_counter


class BeatCounter:
    def __init__(self, tempo=120, quantum=4, start=None):
        if start is None:
            start = system_clock()
        self._start = start
        self._period = 60 / tempo
        self._quantum = quantum
        self._resync_events = set()

    @property
    def tempo(self):
        return 60 / self._period

    @tempo.setter
    def tempo(self, tempo):
        self.set_tempo(tempo)

    @property
    def quantum(self):
        return self._quantum

    @quantum.setter
    def quantum(self, quantum):
        self.set_quantum(quantum)

    @property
    def start(self):
        return self._start

    @property
    def beat(self):
        return self.beat_at_time(system_clock())

    @property
    def phase(self):
        return self.beat_at_time(system_clock()) % self._quantum

    def reset(self, tempo=120, quantum=4, start=None):
        if start is None:
            start = system_clock()
        self._start = start
        self._period = 60 / tempo
        self._quantum = quantum
        self._resync()

    def beat_at_time(self, timestamp):
        return (timestamp - self._start) / self._period

    def phase_at_time(self, timestamp):
        return (timestamp - self._start) / self._period % self._quantum

    def time_at_beat(self, beat):
        return beat * self._period + self._start

    def set_tempo(self, tempo, timestamp=None):
        if timestamp is None:
            timestamp = system_clock()
        period = 60 / tempo
        beat = (timestamp - self._start) / self._period
        self._start = timestamp - beat * period
        self._period = period
        self._resync()

    def set_quantum(self, quantum, timestamp=None):
        if timestamp is None:
            timestamp = system_clock()
        beat = (timestamp - self._start) / self._period
        phase = beat % self._quantum
        self._quantum = quantum
        self.set_phase(phase, timestamp)

    def set_phase(self, phase, timestamp=None, backslip_limit=0):
        if timestamp is None:
            timestamp = system_clock()
        beat = (timestamp - self._start) / self._period
        adjustment = (phase - beat) % self._quantum
        if adjustment > self._quantum - backslip_limit:
            adjustment -= self._quantum
        self._start -= adjustment * self._period
        self._resync()

    def _resync(self):
        for resync_event in self._resync_events:
            resync_event.set()

    async def wait_for_beat(self, beat):
        resync_event = asyncio.Event()
        self._resync_events.add(resync_event)
        try:
            while True:
                timestamp = self.time_at_beat(beat)
                now = system_clock()
                if now >= timestamp:
                    return
                try:
                    await asyncio.wait_for(resync_event.wait(), timeout=timestamp - now)
                except asyncio.TimeoutError:
                    pass
                else:
                    resync_event.clear()
        finally:
            self._resync_events.remove(resync_event)

    def __repr__(self):
        return f"BeatCounter(tempo={60/self._period:.1f}, quantum={self._quantum}, start={self._start:.1f})"


class TapTempo:
    def __init__(self, rounding=2):
        self._rounding = rounding
        self._taps = []

    def __len__(self):
        return len(self._taps)

    def tap(self, timestamp):
        self._taps.append(timestamp)

    def reset(self):
        self._taps = []

    def apply(self, counter, timestamp=None, backslip_limit=0):
        if timestamp is None:
            timestamp = system_clock()
        if self._taps:
            if len(self._taps) > 1:
                tzero, period = polyfit(range(len(self._taps)), self._taps, 1)
                tempo = 60 / period
                if self._rounding:
                    tempo = round(tempo * self._rounding) / self._rounding
                counter.set_tempo(tempo, timestamp)
                counter.set_phase(0, timestamp=tzero, backslip_limit=backslip_limit)
            else:
                counter.set_phase(0, timestamp=self._taps[0], backslip_limit=backslip_limit)
        self._taps = []



"""
Video scene node
"""

from . import Shader, COLOR_FORMATS
from ...cache import SharedCache
from .glconstants import GL_SRGB8
from .glsl import TemplateLoader


class Video(Shader):
    DEFAULT_VERTEX_SOURCE = TemplateLoader.get_template('video.vert')
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('video.frag')

    def __init__(self, glctx):
        super().__init__(glctx)
        self._filename = None
        self._frame0 = None
        self._frame1 = None
        self._frame0_texture = None
        self._frame1_texture = None
        self._colorbits = None

    def release(self):
        self._frame0_texture = None
        self._frame1_texture = None
        self._frame0 = None
        self._frame1 = None
        self._colorbits = None
        super().release()

    @property
    def child_textures(self):
        return {'frame0': self._frame0_texture, 'frame1': self._frame1_texture}

    def similar_to(self, node):
        return super().similar_to(node) and node.get('filename', 1, str) == self._filename

    async def update(self, engine, node, **kwargs):
        references = kwargs.setdefault('references', {})
        if node_id := node.get('id', 1, str):
            references[node_id] = self
        self.hidden = node.get('hidden', 1, bool, False)
        self._filename = node.get('filename', 1, str)
        position = node.get('position', 1, float, 0)
        loop = node.get('loop', 1, bool, False)
        threading = node.get('thread', 1, bool, False)
        if self._filename is not None:
            ratio, frame0, frame1 = SharedCache[self._filename].read_video_frames(self, position, loop, threading=threading)
        else:
            ratio, frame0, frame1 = 0, None, None
        colorbits = node.get('colorbits', 1, int, self.glctx.extra['colorbits'])
        if colorbits not in COLOR_FORMATS:
            colorbits = self.glctx.extra['colorbits']
        if self._texture is not None and (frame0 is None or (self.width, self.height) != (frame0.width, frame0.height)) \
                or colorbits != self._colorbits:
            self.release()
        if frame0 is None:
            return
        if self._texture is None:
            self.width, self.height = frame0.width, frame0.height
            depth = COLOR_FORMATS[colorbits]
            self._texture = self.glctx.texture((self.width, self.height), 4, dtype=depth.moderngl_dtype)
            self._framebuffer = self.glctx.framebuffer(color_attachments=(self._texture,))
            self._frame0_texture = self.glctx.texture((self.width, self.height), 3, internal_format=GL_SRGB8)
            self._frame1_texture = self.glctx.texture((self.width, self.height), 3, internal_format=GL_SRGB8)
            self._colorbits = colorbits
        if frame0 is self._frame1 or frame1 is self._frame0:
            self._frame0_texture, self._frame1_texture = self._frame1_texture, self._frame0_texture
            self._frame0, self._frame1 = self._frame1, self._frame0
        if frame0 is not self._frame0:
            rgb_frame = frame0.to_rgb()
            plane = rgb_frame.planes[0]
            data = rgb_frame.to_ndarray().data if plane.line_size > plane.width * 3 else plane
            self._frame0_texture.write(memoryview(data))
            self._frame0 = frame0
        if frame1 is None:
            self._frame1 = None
        elif frame1 is not self._frame1:
            rgb_frame = frame1.to_rgb()
            plane = rgb_frame.planes[0]
            data = rgb_frame.to_ndarray().data if plane.line_size > plane.width * 3 else plane
            self._frame1_texture.write(memoryview(data))
            self._frame1 = frame1
        interpolate = node.get('interpolate', 1, bool, False)
        self.render(node, ratio=ratio if interpolate else 0, **kwargs)


SCENE_NODE_CLASS = Video
